Package: amregtest
Title: Runs Allelematch Regression Tests
Version: 1.0.3
Authors@R: c(
    person("Department of Wildlife, Fish and Environmental Studies at Swedish University of Agricultural Sciences", role = "cph"),
    person("Göran", "Spong", , "Goran.Spong@slu.se", role = "cph",
           comment = "Senior Lecturer at the Department of Wildlife, Fish and Environmental Studies"),
    person("Paul", "Galpern", , "pgalpern@gmail.com", role = "ctb",
           comment = "Author of the excellent package 'allelematch' from which the five amExample data files have been copied under MIT license"),
    person("Torvald", "Staxler", , "torvald.staxler@telia.com", role = c("aut", "cre"))
  )
Description: Automates regression testing of package 'allelematch'. Over
    2500 tests covers all functions in 'allelematch', reproduces the
    examples from the documentation and includes negative tests. The
    implementation is based on 'testthat'.
License: MIT + file LICENSE
Depends: R (>= 2.10)
Imports: allelematch, digest, remotes, testthat (>= 3.0.0), utils,
        withr
Config/testthat/edition: 3
Encoding: UTF-8
Language: en-US
LazyData: true
RoxygenNote: 7.3.1
NeedsCompilation: no
Packaged: 2024-04-12 13:58:14 UTC; Torva
Author: Department of Wildlife, Fish and Environmental Studies at Swedish
    University of Agricultural Sciences [cph],
  Göran Spong [cph] (Senior Lecturer at the Department of Wildlife, Fish
    and Environmental Studies),
  Paul Galpern [ctb] (Author of the excellent package 'allelematch' from
    which the five amExample data files have been copied under MIT
    license),
  Torvald Staxler [aut, cre]
Maintainer: Torvald Staxler <torvald.staxler@telia.com>
Repository: CRAN
Date/Publication: 2024-04-12 15:30:09 UTC
Built: R 4.2.3; ; 2024-04-12 18:36:52 UTC; unix

Package: amregtest
Title: Runs Allelematch Regression Tests
Version: 1.1.0
Date: 2026-07-04
Authors@R: c(
    person("Department of Wildlife, Fish and Environmental Studies at Swedish University of Agricultural Sciences",
           role = "cph"),
    person("Göran", "Spong", email = "Goran.Spong@slu.se", role = "cph",
           comment = "Senior Lecturer at the Department of Wildlife, Fish and Environmental Studies"),
    person("Paul", "Galpern", email = "pgalpern@gmail.com", role = "ctb",
           comment = "Author of the excellent package 'allelematch' from which the five amExample data files have been copied under MIT license"),
    person("Torvald", "Staxler", email = "torvald.staxler@telia.com", role = c("aut", "cre")))
Description: Automates regression testing of package 'allelematch'. Over
    2500 tests cover all functions in 'allelematch', reproduce the
    examples from the documentation, and include negative tests. The
    implementation is based on 'testthat'.
License: MIT + file LICENSE
URL: https://github.com/torstax/amregtest
BugReports: https://github.com/torstax/amregtest/issues
Depends: R (>= 3.5), allelematch
Imports: digest, R6, remotes, testthat (>= 3.0.0), utils, withr
Config/testthat/edition: 3
Encoding: UTF-8
Language: en-US
LazyData: true
Config/roxygen2/markdown: TRUE
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-07-04 19:13:38 UTC; Torva
Author: Department of Wildlife, Fish and Environmental Studies at Swedish
    University of Agricultural Sciences [cph],
  Göran Spong [cph] (Senior Lecturer at the Department of Wildlife, Fish
    and Environmental Studies),
  Paul Galpern [ctb] (Author of the excellent package 'allelematch' from
    which the five amExample data files have been copied under MIT
    license),
  Torvald Staxler [aut, cre]
Maintainer: Torvald Staxler <torvald.staxler@telia.com>
Repository: CRAN
Date/Publication: 2026-07-04 23:10:14 UTC
Built: R 4.5.2; ; 2026-07-05 02:52:11 UTC; unix

# Generated by roxygen2: do not edit by hand

export(artList)
export(artRun)
export(artVersion)
import(allelematch)
import(digest)
import(remotes)
import(testthat)
import(utils)
import(withr)
importFrom(R6,R6Class)

# amregtest

<!-- badges: start -->

<!-- badges: end -->

“`amregtest`” stands for “`allelematch` regression test”.
[`allelematch`](https://cran.r-project.org/package=allelematch) is an R
package for identifying and removing duplicate multilocus genotypes from
microsatellite data sets. `amregtest` verifies that `allelematch`
remains backwards compatible across versions.

By decoupling the regression test suite from the main `allelematch`
source code and its specific `testthat` dependencies, `amregtest` allows
you to run regression tests against *any* version of `allelematch`
without modifying the package under test.

## Installation

You can install the development version of `amregtest` from GitHub with:

``` r
# install.packages("devtools")
devtools::install_github("torstax/amregtest")
```

## Usage

``` r
library(allelematch)
library(amregtest)
```

Run the full regression test suite against your currently installed
version of `allelematch` (takes a few minutes):

``` r
artRun()
```

    ##     About to test installed version of allelematch:  <<<2.5.5>>>
    ## FAILED  WARNINGS  SKIPPED  OK      FILE
    ## ------------------------------------------------------------------------ 
    ## 0       0         0        74      test-allelematch_1-amDataset.R
    ## 0       0         0        10      test-allelematch_1-amDataset_negative.R
    ## 0       0         0        30      test-allelematch_2-amMatrix.R
    ## 0       0         0        20      test-allelematch_2-amMatrix_negative.R
    ## 0       0         0        362     test-allelematch_3-amPairwise.R
    ## 0       0         0        39      test-allelematch_3-amPairwise_negative.R
    ## 0       1         0        19      test-allelematch_3-amPairwise_print.R
    ## 0       0         0        1344    test-allelematch_4-amCluster.R
    ## 0       1         0        19      test-allelematch_4-amCluster_print.R
    ## 0       0         0        22      test-allelematch_5-amAlleleFreq.R
    ## 0       0         0        768     test-allelematch_6-amUnique.R
    ## 0       0         0        42      test-allelematch_6-amUnique_negative.R
    ## 0       1         0        21      test-allelematch_6-amUnique_print.R
    ## 0       0         0        13      test-allelematch_7-amUniqueProfile.R
    ## 0       0         0        16      test-amExample1.R
    ## 0       0         0        14      test-amExample2.R
    ## 0       0         0        14      test-amExample3.R
    ## 0       0         0        14      test-amExample4.R
    ## 0       0         0        4       test-art_api.R
    ## 0       0         0        8       test-ggData.R
    ## ------------------------------------------------------------------------ 
    ## 0       3         0        2853    TOTAL
    ##     Done testing installed version of allelematch:  <<<2.5.5>>>

To run only tests whose names match a Perl regular expression:

``` r
artRun(filter = "amExample2|amExample4")
```

    ##     About to test installed version of allelematch:  <<<2.5.5>>>
    ## FAILED  WARNINGS  SKIPPED  OK      FILE
    ## ------------------------------------------------------------------------ 
    ## 0       0         1        7       test-amExample2.R
    ## 0       0         1        9       test-amExample4.R
    ## ------------------------------------------------------------------------ 
    ## 0       0         2        16      TOTAL
    ##     Done testing installed version of allelematch:  <<<2.5.5>>>

List all available tests without running them:

``` r
artList()
```

    ## 
    ## Tests in files under "C:/Users/Torva/AppData/Local/Temp/RtmpEf1k0b/temp_libpath53f819d62beb/amregtest/tests/testthat/":
    ## 
    ## Tests by functions in allelematch:
    ##  [1] "allelematch_1-amDataset"          
    ##  [2] "allelematch_1-amDataset_negative" 
    ##  [3] "allelematch_2-amMatrix"           
    ##  [4] "allelematch_2-amMatrix_negative"  
    ##  [5] "allelematch_3-amPairwise"         
    ##  [6] "allelematch_3-amPairwise_negative"
    ##  [7] "allelematch_3-amPairwise_print"   
    ##  [8] "allelematch_4-amCluster"          
    ##  [9] "allelematch_4-amCluster_print"    
    ## [10] "allelematch_5-amAlleleFreq"       
    ## [11] "allelematch_6-amUnique"           
    ## [12] "allelematch_6-amUnique_negative"  
    ## [13] "allelematch_6-amUnique_print"     
    ## [14] "allelematch_7-amUniqueProfile"    
    ## 
    ## Reproduction of the examples in 'allelematchSuppDoc.pdf':
    ## [1] "amExample1" "amExample2" "amExample3" "amExample4"
    ## 
    ## Other:
    ## [1] "art_api" "ggData"

Check the installed versions of `allelematch` and `amregtest`:

``` r
artVersion()
```

    ## 
    ##     Version of package 'amregtest' is 1.0.10
    ##     Installed (and thus tested) version of package 'allelematch' is: 2.5.5

# A plain-text, one-line-per-file testthat reporter.
# Prints a header and one line per test file with columns:
# FAILED, WARNINGS, SKIPPED, OK, FILE.
# No spinner, no ANSI sequences -- suitable for knitr / non-interactive output.
#
#' @noRd
ColumnReporter <- R6::R6Class("ColumnReporter",
  inherit = testthat::Reporter,
  public = list(

    # Per-file state
    .file_name = "",
    .n_fail    = 0L,
    .n_warn    = 0L,
    .n_skip    = 0L,
    .n_ok      = 0L,

    # Overall totals (for end_reporter summary)
    .total_fail = 0L,
    .total_warn = 0L,
    .total_skip = 0L,
    .total_ok   = 0L,

    start_reporter = function() {
      cat(sprintf("%-6s  %-8s  %-7s  %-6s  %s\n",
                  "FAILED", "WARNINGS", "SKIPPED", "OK", "FILE"))
      cat(strrep("-", 72), "\n")
    },

    start_file = function(filename) {
      self$.file_name <- filename
      self$.n_fail    <- 0L
      self$.n_warn    <- 0L
      self$.n_skip    <- 0L
      self$.n_ok      <- 0L
    },

    add_result = function(context, test, result) {
      if (inherits(result, c("expectation_failure", "expectation_error"))) {
        self$.n_fail <- self$.n_fail + 1L
      } else if (inherits(result, "expectation_warning")) {
        self$.n_warn <- self$.n_warn + 1L
      } else if (inherits(result, "expectation_skip")) {
        self$.n_skip <- self$.n_skip + 1L
      } else {
        self$.n_ok <- self$.n_ok + 1L
      }
    },

    end_file = function() {
      cat(sprintf("%-6d  %-8d  %-7d  %-6d  %s\n",
                  self$.n_fail, self$.n_warn, self$.n_skip, self$.n_ok,
                  self$.file_name))

      self$.total_fail <- self$.total_fail + self$.n_fail
      self$.total_warn <- self$.total_warn + self$.n_warn
      self$.total_skip <- self$.total_skip + self$.n_skip
      self$.total_ok   <- self$.total_ok   + self$.n_ok
    },

    end_reporter = function() {
      cat(strrep("-", 72), "\n")
      cat(sprintf("%-6d  %-8d  %-7d  %-6d  %s\n",
                  self$.total_fail, self$.total_warn, self$.total_skip,
                  self$.total_ok, "TOTAL"))
    }
  )
)

###############################################################################
### allelematch test engine ###
###############################################################################

#' Returns package version
#'
#' @description
#' Displays version of this package (\link{amregtest}) and of [allelematch][allelematch::allelematch-package],
#' together with build timestamps.\cr
#' \cr
#' The version is specified in the file DESCRIPTION, tag "Version: ".
#'
#'
#' @param verbose logical. If TRUE (the default), prints additional info to stdout,
#' including versions and build timestamps of 'allelematch' and 'amregtest'.
#'
#' @return The loaded version of this package (\link{amregtest-package}) in a character vector of length one
#'
#' @examples
#' # See what version of packages 'allelematch' and 'amregtest'
#' # are currently loaded:
#' artVersion()
#'
#' # List the available tests:
#' artList()
#' \donttest{
#' # Run all the tests:
#' # artRun()  # Takes several minutes
#'
#' # Run the first of the available tests:
#' artRun(filter="allelematch_1-amDataset$")
#' }
#'
#'
#' @seealso \link{artList}, \link{artRun} and \link{amregtest}
#' @export
artVersion <- function(verbose=TRUE) {
    stopifnot(is.logical(verbose))

    installedArtVersion = toString(utils::packageVersion("amregtest"))
    installedAmVersion  = toString(utils::packageVersion("allelematch"))

    if (verbose) {
        cat("\n    Version of package 'amregtest' is", installedArtVersion)
        cat("\n    Installed (and thus tested) version of package 'allelematch' is:", installedAmVersion)
    }
    return(invisible(installedArtVersion))
}


#' Lists available tests in `amregtest` without running them
#'
#' @description
#' Use the output to select a value for parameter `filter` to \link{artRun}.
#' Useful when debugging.
#'
#' @param verbose logical. If TRUE (the default), prints additional info to stdout
#'
#' @return A character vector containing the names of all the tests
#'
#' @examples
#' # See what version of packages 'allelematch' and 'amregtest'
#' # are currently loaded:
#' artVersion()
#'
#' # List the available tests:
#' artList()
#' \donttest{
#' # Run all the tests:
#' # artRun()  # Takes several minutes
#'
#' # Run the first of the available tests:
#' artRun(filter="allelematch_1-amDataset$")
#' }
#'
#' @seealso \link{artVersion} and \link{artRun}
#'
#' @export
artList <- function(verbose=TRUE) {
    stopifnot(is.logical(verbose))

    root = paste(system.file(package = "amregtest"), "tests/testthat/", sep="/")

    all = gsub("^test-(.+?)\\.R", "\\1", grep("^test-.+?\\.R", list.files(root), value=TRUE), perl=TRUE)

    if (verbose) {
        cat('\nTests in files under "', root, '":\n', sep="")

        cat("\nTests by functions in allelematch:\n")
        print(grep("^allelematch", all, value=TRUE, perl=TRUE), width=50)

        cat("\nReproduction of the examples in 'allelematchSuppDoc.pdf':\n")
        print(grep("^amExample", all, value=TRUE, perl=TRUE))

        cat("\nOther:\n")
        print(grep("^allelematch|^amExample", all, value=TRUE, invert=TRUE, perl=TRUE))
        cat("\n")
    }

    return(invisible(all))
}


#' Runs the regression test
#'
#' @description
#' Runs regression tests on package [allelematch][allelematch::allelematch-package] to make sure it is backwards compatible.\cr
#' \cr
#' The full set of tests will take a couple of minutes. \cr
#' \cr
#' Call \link{artList} to see the available tests without running them.
#'
#' @return A list (invisibly) containing data about the test results as returned by [testthat::test_package]
#'
#' @details
#' If any of the test executed with \link{artRun} should fail, then we want to be able
#' to run that specific test under the debugger.\cr
#' \cr
#' Set a breakpoint in `allelematch.R` and call `artRun(filter="<the test that reproduces the problem>")`\cr
#' \cr
#' Note that it is the last loaded version of `allelematch` that will be executed,
#' not the last edited. In RStudio, CTRL+SHIFT+B will build, install and load.
#'
#'
#' @param filter    If specified, only tests with names matching this perl regular
#'                  expression will be executed. Character vector of length 1. See also \link{artList}
#' @param verbose   logical. If TRUE (the default), prints version of tested allelematch to stdout
#'
#' @examples
#' # See what version of packages 'allelematch' and 'amregtest'
#' # are currently loaded:
#' artVersion()
#'
#' # List the available tests:
#' artList()
#' \donttest{
#' # Run all the tests:
#' # artRun()  # Takes several minutes
#'
#' # Run the first of the available tests:
#' artRun(filter="allelematch_1-amDataset$")
#' }
#'
#' @seealso \link{artVersion} and \link{artList}
#'
#' @export
artRun <- function(filter="", verbose=TRUE) {
    stopifnot(is.character(filter) && length(filter)==1)
    stopifnot(is.logical(verbose))

    # Propagate the verbosity to the deferred cleanup code in helper-temp-leaks.R
    # withr ensures the effect lasts until the ENTIRE function returns
    env_value = if(verbose) "true" else "false"
    withr::local_envvar(c("ART_VERBOSE" = env_value))

    installedVersion = toString(utils::packageVersion("allelematch"))
    if (verbose) cat("    About to test installed version of allelematch:  <<<", installedVersion, ">>>\n", sep="")

    reporter <-
        if (!verbose) testthat::check_reporter() # Keep silent unless there are errors.
        else if(interactive()) "Progress"        # Nice interactive progress bar in RStudio
        else ColumnReporter$new()                # Nice columns to stdout

    result = list()
    if (filter != "^$") result = testthat::test_package("amregtest", reporter=reporter , filter=filter) # We can't start tests recursively, even for coverage tests
    if (verbose) cat("    Done testing installed version of allelematch:  <<<", installedVersion, ">>>\n", sep="")
    return(invisible(result))
}

test_that("We are running the 3rd edition of testthat", code = {
  # If this tests fails, then call
  #   usethis::use_testthat(3)
  # to configure DESCRIPTION to use 3rd edition of 'testthat'.
  testthat::expect_gte(!!testthat::edition_get(), 3)
})

test_that("art_api functions working", {

  sink(nullfile())

  # These tests should give 100% test coverage for our API calls:
  expect_no_error(artVersion())
  expect_output(artList(), "Reproduction of the examples in 'allelematchSuppDoc.pdf':")
  expect_no_error(artRun("^$")) # Special filter to avoid starting tests recursively

  sink()
})

# 'tdd-*.R' files are not picked up automatically by 'testthat',
# but can be executed manually.
#
# This file tests the art machinery.


test_that("We can read environment set from artRun()", code = {

    art_env = list(
        ART_CALLERS_WD = getwd(),
        ART_GENERATE_HTML = as.character(FALSE)
    )
    assign("art_env", art_env, globalenv())

    artEnv <- new.env(parent=globalenv())
    attr(artEnv, "name") = "artEnv"

    artEnv2 <- new.env(parent=globalenv())
    attr(artEnv2, "name") = "artEnv"

    testthat::expect_identical(environmentName(artEnv2), environmentName(artEnv))

    assign("ART_CALLERS_WD", getwd(), artEnv)

    # dir = Sys.getenv("ART_CALLERS_WD")

    # testthat::expect_identical(get("ART_CALLERS_WD", artEnv), "hemma hos nisse")
    testthat::expect_identical(get("art_env", globalenv())$ART_CALLERS_WD, "hemma hos nisse")
})


#-
#- @details
#- TODO : Move this to package introduction!
#-
#- NOTE that the output of [sort] is platform dependent.
#- You get different results based on the locale. And [allelematch] uses [sort].\cr\cr
#-
#- The sort order also depends on the size of the data to be sorted.
#- See the description of "radix" under [sort].\cr\cr
#-
#- [testthat] (and R CMD check) makes sure that the tests behave the same way
#- on every platform, by setting the collation locale to "C" and the language to "en".\cr
#- See github issue
#- \href{https://github.com/r-lib/testthat/issues/1181#issuecomment-692851342}{locale / collation used in testthat #1181}\cr\cr

#' @name artData
#'
#' @title Example data used by \link{amregtest}
#'
#' @description
#' This example data is used when testing allelematch backwards compatibility
#' using \link{artRun}. The tests load this data and passes it to [amDataset][allelematch::amDataset].\cr
#' \cr
#' It includes data that was imported from version 5.2.1 of [allelematch][allelematch::allelematch-package].
#' It was still unchanged in 5.2.5. \cr
#' \tabular{clcl}{
#'  `  ` \tab \link{amExample1}`  `\tab `  ` \tab Example 1 High quality data set\cr
#'  `  ` \tab \link{amExample2}\tab \tab Example 2 Good quality data set\cr
#'  `  ` \tab \link{amExample3}\tab \tab Example 3 Marginal quality data set\cr
#'  `  ` \tab \link{amExample4}\tab \tab Example 4 Low quality data set\cr
#'  `  ` \tab \link{amExample5}\tab \tab Example 5 Wildlife data set\cr
#' }
#' See \href{https://github.com/cran/allelematch/blob/2.5.1/inst/doc/allelematchSuppDoc.pdf}{allelematchSuppDoc.pdf}
#' for a more detailed description. \cr
#'
#' It also includes a large data set gathered from field work:\cr
#' \tabular{clcl}{
#'  `  ` \tab \link{ggSample}`  `\tab `  ` \tab Very large wildlife data set\cr
#' }
#'
#' @format Data frames with varying numbers of samples in rows, and alleles in columns. Missing data is represented as "-99".
#'
#' @docType data
#' @references \url{https://github.com/cran/allelematch}
#' @references \href{https://github.com/cran/allelematch/blob/2.5.1/inst/doc/allelematchSuppDoc.pdf}{allelematchSuppDoc.pdf}
#' @keywords data
NULL

#' Example 1 High quality data set
#'
#' This is sample data copied from [amExampleData][allelematch::amExampleData] in version 5.2.1
#' of package [allelematch][allelematch::allelematch-package]. We use this data to test 'allelematch' backwards compatibility.
#'
#' The data in this example is simulated to represent a high quality data set that might
#' result from a laboratory protocol where samples were run multiple times to confirm their
#' identity. It has no genotyping error, a near-zero missing data load, and approximately
#' 60% of the individuals have been artificially resampled more than once.
#'
#' @format Data frame with samples in rows, and alleles in columns. Missing data is represented as "-99".
#'
#' @name amExample1
#' @docType data
#' @references \url{https://github.com/cran/allelematch}
#' @keywords data
NULL

#' Example 2 Good quality data set
#'
#' This is sample data copied from [amExampleData][allelematch::amExampleData] in version 5.2.1
#' of package [allelematch][allelematch::allelematch-package].
#'
#' The data in this example have also been simulated, this time to reflect the qualities
#' of good quality data set, where genotyping error and missing data exist, but these can
#' be confidently handled by allelematch without manual intervention. At each locus a
#' random 4% of heterozygotes lost their second allele to simulate an allele dropout, and
#' a random 4% of samples at each locus had alleles set to missing.
#'
#' @format Data frame with samples in rows, and alleles in columns. Missing data is represented as "-99".
#'
#' @name amExample2
#' @docType data
#' @references \url{https://github.com/cran/allelematch}
#' @keywords data
NULL

#' Example 3 Marginal quality data set
#'
#' This is sample data copied from [amExampleData][allelematch::amExampleData] in version 5.2.1
#' of package [allelematch][allelematch::allelematch-package].
#'
#' The data in this example have been simulated to represent a data set of marginal
#' quality where the use of allelematch combined with careful manual review of the
#' results is required to achieve a confident assessment of the unique genotypes. At each
#' locus a random 4% of heterozygotes lost their second allele to simulate an allele dropout,
#' and a random 10% of samples at each locus had alleles set to missing.
#'
#' @format Data frame with samples in rows, and alleles in columns. Missing data is represented as "-99".
#'
#' @name amExample3
#' @docType data
#' @references \url{https://github.com/cran/allelematch}
#' @keywords data
NULL

#' Example 4 Low quality data set
#'
#' This is sample data copied from [amExampleData][allelematch::amExampleData] in version 5.2.1
#' of package [allelematch][allelematch::allelematch-package].
#'
#' For this example we have simulated a low quality data set where uncertainty created
#' by genotyping error and missing data, combined with a lack of information in the form
#' of allelic diversity across loci will result in a low confidence assessment of the unique
#' genotypes. At each locus a random 6% of heterozygotes lost their second allele to
#' simulate an allele dropout, and a random 20% of samples at each locus had alleles set
#' to missing.
#'
#' @format Data frame with samples in rows, and alleles in columns. Missing data is represented as "-99".
#'
#' @name amExample4
#' @docType data
#' @references \url{https://github.com/cran/allelematch}
#' @keywords data
NULL

#' Example 5 Wildlife data set
#'
#' This is sample data copied from [amExampleData][allelematch::amExampleData] in version 5.2.1
#' of package [allelematch][allelematch::allelematch-package].
#'
#' In this final example we use real data from the non-invasive sampling of a wildlife
#' population. The data have been anonymized by changing sampling details. A single
#' column giving the gender is also available and we show how this can be used as an extra
#' locus. Missing data is also more common at some loci than at others, with a total load
#' of about 10%.
#'
#' @format Data frame with samples in rows, and alleles in columns. Missing data is represented as "-99".
#'
#' @name amExample5
#' @docType data
#' @references \url{https://github.com/cran/allelematch}
#' @keywords data
NULL


#' @name ggSample
#'
#' @title Data sets originating from GG work
#'
#' @description
#' Large data set gathered from field work in 2022. Here used to test package [allelematch][allelematch::allelematch-package] for backwards compatibility.
#' Combines a reference db of known individuals with new samples to be analyzed.\cr\cr
#'
#' @format This data is saved on semicolon (";") delimited .csv format, as described under 'Details' in [utils::data].\cr
#' Data samples in rows, and alleles in columns. Missing data is represented as "-99".
#'
#' @references \url{https://github.com/cran/allelematch}
#'
#' @docType data
#' @keywords data
#' @keywords internal
NULL

# Deferred TEMP-leak cleanup — runs once after the entire test suite.
#
# This must live in a setup-*.R file, not a helper-*.R file.
# Helper files are sourced during load_all(), before the test session exists,
# so testthat::teardown_env() is not yet initialised at that point.
# Setup files are sourced after the session is ready.
#
# The .art_leaked collector and the test_that wrapper are in helper-temp-leaks.R.

withr::defer(
  {
    keep_files <- toupper(Sys.getenv("ART_KEEP_LEAKED_FILES_IN_TEMP")) == "TRUE"
    if (length(.art_leaked) > 0) {
      # if (!interactive()) graphics.off()
      #
      # is_png <- grepl("\\.png$", .art_leaked, ignore.case = TRUE)
      # keep   <- if (interactive()) .art_leaked[ is_png] else character(0)
      # to_del <- if (interactive()) .art_leaked[!is_png] else .art_leaked

      graphics.off()
      keep   <- if (keep_files) .art_leaked  else character(0)
      to_del <- if (keep_files) character(0) else .art_leaked

      if (length(to_del) > 0) {
        suppressWarnings(file.remove(to_del))
        message(sprintf(
          "Cleaning up %d leaked TEMP file(s):\n  %s",
          length(to_del),
          paste(to_del, collapse = "\n  ")
        ))
      }
      if (length(keep) > 0) {
        message(sprintf(
          "Keeping %d file(s) for inspection:\n  %s",
          length(keep),
          paste(keep, collapse = "\n  ")
        ))
      }
    }
  },
  envir = testthat::teardown_env()
)

#library(amregtest)
require(amregtest, quietly = TRUE)

# See https://stackoverflow.com/questions/74379129/r-using-arguments-passed-to-for-named-arguments-when-arguments-are-subst
options(warnPartialMatchArgs=TRUE)

# NOTE that the output of [sort] is platform dependent.
# You get different results based on the locale. And [allelematch] uses [sort].
#
# The sort order also depends on the size of the data to be sorted.
# See the description of "radix" under [sort].
#
# [testthat] (and R CMD check) makes sure that the tests behave the same way
# on every platform, by setting the collation locale to "C" and the language to "en".
# See issue "locale / collation used in testthat #1181"
#  at https://github.com/r-lib/testthat/issues/1181#issuecomment-692851342

######################################
### Helper functions for the tests ###
######################################

# Load the amregtest version of example data sets from this package
# into the global environment. Needs to match the tests:
data("amExample1", package = "amregtest", envir = .GlobalEnv)
data("amExample2", package = "amregtest", envir = .GlobalEnv)
data("amExample3", package = "amregtest", envir = .GlobalEnv)
data("amExample4", package = "amregtest", envir = .GlobalEnv)
data("amExample5", package = "amregtest", envir = .GlobalEnv)

# Turns `...` into a string of <name>=<value> pairs
helpArgToString <- function(...) {
  # form the `...` arguments into a string on the form "<name1>=<value1>, <name2>=<value2> ..."
  v = character(...length()) # Reserve space for a vector of strings

  # Paste together <name>=<value> pairs.
  # the <name> comes from `...names()[n]`. The <value> comes from `...elt(n)`
  # for(n in 1:...length()) v[n] = paste(...names()[n], format(...elt(n)), sep="=")
  for(n in 1:...length()) v[n] = paste(...names()[n], deparse(...elt(n)), sep="=")

  # Remove spurious empty pairs from the vector:
  v = v[! v == "=NULL"]

  # Return a string were the pairs are ", " -separated:
  ret = paste(v, collapse=", ")

  return(ret)
}

# Attempts to prepend the supplied file name with a directory
# where the user can find the generated .html files:
helpHtml <- function(file) {
    dir = Sys.getenv("ART_CALLERS_WD")
    if(dir == "") {
        dir = ifelse( grepl("/tests/testthat$", getwd()), "../..", ".") # testthat changes getwd() to tests/testthat/.
        dir = normalizePath(dir, winslash = "/")
    }
    # dir = sub("^(C):", "/\\1", dir, perl=TRUE, fixed=FALSE)
    dir = sub("^C:", "", dir, perl=TRUE, fixed=FALSE)
    if(!dir.exists(dir)) stop("\n    dir = '", dir, "' does not exist!\n    getwd() = '", getwd(), "' ", sep="")
    longfile = paste(dir, "/", file, sep="")
    cat("\n    Writing html to :", longfile)

    return(longfile)
}

test_that("Loop the Loop", {
  
  # We want to run amPairwise many times with many combinations of parameters
  # and we want to compare the results with previous runs. Like this:
  snapshot_amPairwise <- function(ds, ...) {
    
    # Log the call to the snapshot file:
    argstr = helpArgToString(...)
    cmdstr = paste("amPairwise(", ds, ", ", argstr, ")", sep="") ; expect_snapshot(cat(cmdstr))
    
    # Make the call:
    pw <- amPairwise(amDatasetFocal=get(ds), ...)
    
    # Log the result to the snapshot file:
    # expect_snapshot(summary.amPairwise(pw)) # Too big. :-(
    expect_snapshot_value(pw, style = "json2")
    
    return(pw)
  }
  
  # Run different data sets with different qualities through the same loops:
  miniExample = data.frame(
    "LOC1a"         = c(11:14),
    "LOC1b"         = c(21:24),
    "LOC2a"         = c(31:33, -99),
    "LOC2b"         = c(41:44)
  )
  data("amExample1")
  data("amExample2") ; amExample2 = amExample2[c(1:20),] # Just keep the first 20 rows to save speed and disk
  data("amExample3") ; amExample3 = amExample3[c(1:20),] # Just keep the first 20 rows to save speed and disk
  data("amExample4") ; amExample4 = amExample4[c(1:20),] # Just keep the first 20 rows to save speed and disk
  data("amExample5") ; amExample5 = amExample5[c(1:20),] # Just keep the first 20 rows to save speed and disk

  amdataMini     = amDataset(miniExample)
  amdataExample1 = amDataset(amExample1, indexColumn="sampleId", metaDataColumn="knownIndividual")
  amdataExample2 = amDataset(amExample2, indexColumn="sampleId", metaDataColumn="knownIndividual")
  amdataExample3 = amDataset(amExample3, indexColumn="sampleId", metaDataColumn="knownIndividual")
  amdataExample4 = amDataset(amExample4, indexColumn="sampleId", metaDataColumn="knownIndividual")
  amdataExample5 = amDataset(amExample5, indexColumn="sampleId", ignoreColumn=c("samplingData", "gender"))

  # Here comes the loops:
  for (amds1 in c("amdataMini", "amdataExample1", "amdataExample2", "amdataExample3", "amdataExample4", "amdataExample5")) {
    #for (amds2 in c("amdataMini", "amdataExample1", "amdataExample2")) { # Must have data columns where all column names are the same
      for (mm in c(0, 1, 3, 5, 7, 9, 11, NA)) {
        for (th in c(NA, 0, 0.1, 0.3, 0.5, 0.7, 0.9, 0.95, 1)) {
          for (meth in c(1, 2)) {
            # Either alleleMismatch or matchThreshold shall be set. Not both:
            if (is.na(mm) == is.na(th)) next
            
            if (!is.na(mm)) {
              snapshot_amPairwise(amds1,  #amds2, 
                       alleleMismatch =  mm,    missingMethod =  meth)
            } else if (!is.na(th)) {
              snapshot_amPairwise(amds1, #amds2,
                       matchThreshold =  th,    missingMethod =  meth)
            } else {
              stop("Unexpected combination of mm=", mm, "and th=", th)
            }
          }
        }
      }
    #}
  }
})


test_that("Value from amPairwise() remains stable", {
  
  # data(amExample1)
  # expect_equal(!!dim(amExample1), c(20,22))
  # 
  # amdata = amDataset1_1= amDataset(amExample1, indexColumn = "sampleId", metaDataColumn = "knownIndividual")
  # expect_snapshot_value(amdata, style = "json2")
  
  # Create valid miniature input sample:    
  sample = miniExample = data.frame(
    "LOC1a"         = c(11:14),
    "LOC1b"         = c(21:24),
    "LOC2a"         = c(31:33, -99),
    "LOC2b"         = c(41:44)
  )
  
  # Remember how the pairwise is calculated for regression testing:
  {
    pw = amPairwise1 = amPairwise(amDataset(sample), matchThreshold=0.95)
    expect_snapshot_value(pw, style = "json2")
    
    expect_snapshot(
      list(
        pw1 = amPairwise(amDataset(sample), matchThreshold=0.95, missingMethod = 1),
        pw2 = amPairwise(amDataset(sample), matchThreshold=0.95, missingMethod = 2)
      )
    )
  }
  
  
  # # Detect changes in the calculation of an amPairwise
  # # whilst regression testing:
  # data("amExample2") # Good quality data set
  # amdataExample2 <- amDataset(amExample2, indexColumn="sampleId",
  #                             metaDataColumn="knownIndividual", missingCode="-99")
  # {
  #   expect_snapshot(pw21 <- amPairwise(amdataExample2, 1))
  #   expect_snapshot_value(pw21, style = "json2")
  #   
  #   expect_snapshot(pw22 <- amPairwise(amdataExample2, 2))
  #   expect_snapshot_value(pw22, style = "json2")
  # }
  
  ## The following amPairwise:es get too big for expect_snapshot_value :-( 
  # data("amExample3") # Marginal quality data set
  # amdataExample3 <- amDataset(amExample3, indexColumn="sampleId",
  #                             metaDataColumn="knownIndividual", missingCode="-99")
  # {
  #   expect_snapshot(pw31 <- amPairwise(amdataExample3, missingMethod = 1))
  #   expect_snapshot_value(pw31, style = "json2")
  # }
  # 
  # data("amExample4") # Poor quality example
  # amdataExample4 <- amDataset(amExample4, indexColumn="sampleId",
  #                             metaDataColumn="knownIndividual", missingCode="-99")
  # {
  #   expect_snapshot(pw41 <- amPairwise(amdataExample4, 1))
  #   expect_snapshot_value(pw41, style = "json2")
  # }
  # 
  # data("amExample5") # Wildlife example
  # amdataExample5 <- amDataset(amExample5, indexColumn="sampleId",
  #                             metaDataColumn="samplingData", missingCode="-99")
  # {
  #   expect_snapshot(pw51 <- amPairwise(amdataExample5, missingMethod = 1))
  #   expect_snapshot_value(pw51, style = "json2")
  # }
  
})

test_that("We are running the 3rd edition of testthat", code = {
    testthat::expect_gte(!!testthat::edition_get(), 3)
})

HTML=isTRUE(Sys.getenv("ART_GENERATE_HTML") == "TRUE") # Set with Sys.setenv(ART_GENERATE_HTML = "TRUE")

overwrite = FALSE # Use TRUE when creating new tests that need new *_expected data.

test_that("amExample4 results from pg 14 in allelematchSuppDoc.pdf are 2.5.3 compatible", code = {

    # Prepare for printing large snapshot files:
    withr::local_options(width=200) # Allow longer lines for the summaries:

    # Follow the instructions from allelematchSuppDoc.pdf, pg 14:
    data("amExample4")
    example4 <- amDataset(amExample4, indexColumn="sampleId",
      metaDataColumn="knownIndividual", missingCode="-99")
    {
        # Ensure that the result is still the same as that from 2.5.3:
        expect_snapshot_value(example4, style="deparse")
    }

    output = capture.output(
        amUniqueProfile(example4, doPlot=TRUE)
    )
    {
        # Ensure that the result is still the same as that from 2.5.3:
        # cat("\nOutput from amUniqueProfile:\n", output, "\nEnd of output\n", sep="\n    ")

        testthat::expect_match(output, "missing data load for input dataset is 0.199 ", fixed=TRUE, all=FALSE)
        testthat::expect_match(output, "allelic diversity for input dataset is 4.8 ", fixed=TRUE, all=FALSE)
        testthat::expect_match(output, "Best guess for optimal parameter at alleleMismatch=1 OR matchThreshold=0.95 OR cutHeight=0.05$", all=FALSE)
        testthat::expect_match(output, "Best guess for unique profile morphology: NoSecondMinimum$", all=FALSE)
        testthat::expect_match(output, "Use extra caution.", fixed=TRUE, all=FALSE)
    }

    output = capture.output(
        uniqueExample4 <- amUnique(example4, alleleMismatch=1)
    )
    {
        testthat::expect_match(output, "assuming genotype columns are in pairs, representing 10 loci$", all=FALSE)

        # Ensure that the result is still the same as that from 2.5.3:
        expect_snapshot_value(uniqueExample4, style="deparse")

        # Generate a summary file:
        summary.amUnique(uniqueExample4, csv=summaryFile <- tempfile("example4_1.csv"))

        # Ensure that the result is still the same as that from 2.5.3
        expect_snapshot_value(read.csv(file=summaryFile, colClasses="character"), style="deparse")
        file.remove(summaryFile)
    }

    if(HTML) {
        summary.amUnique(uniqueExample4, html=helpHtml("example4_1.html"))
    }

    output = capture.output(
        uniqueExample4ballpark <- amUnique(example4, alleleMismatch=6)
    )
    {
        testthat::expect_match(output, "assuming genotype columns are in pairs, representing 10 loci$", all=FALSE)

        # Ensure that the result is still the same as that from 2.5.3:
        # expect_snapshot_value(uniqueExample4ballpark, style="deparse") # (Couldn't serialize)

        # Generate a summary file:
        summary.amUnique(uniqueExample4ballpark, csv=summaryFile <- tempfile("example4_2.csv"))

        # Ensure that the result is still the same as that from 2.5.3
        expect_snapshot_value(read.csv(file=summaryFile, colClasses="character"), style="deparse")
        file.remove(summaryFile)
    }

    if(HTML) {
        summary.amUnique(uniqueExample4ballpark, html=helpHtml("example4_2.html"))
    }

    output = capture.output(
        uniqueExample4high <- amUnique(example4, alleleMismatch=6)
    )
    {
        testthat::expect_match(output, "assuming genotype columns are in pairs, representing 10 loci$", all=FALSE)

        # Ensure that the result is still the same as that from 2.5.3:
        # expect_snapshot_value(uniqueExample4high, style="deparse") # Couldn't serialize

        # Generate a summary file:
        summary.amUnique(uniqueExample4high, csv=summaryFile <- tempfile("example4_3.csv"))

        # Ensure that the result is still the same as that from 2.5.3
        expect_snapshot_value(read.csv(file=summaryFile, colClasses="character"), style="deparse")
        file.remove(summaryFile)
    }

    if(HTML) {
        summary.amUnique(uniqueExample4high, html=helpHtml("example4_3.html"))
    }


})

test_that("We are running the 3rd edition of testthat", code = {
    testthat::expect_gte(!!testthat::edition_get(), 3)
})

HTML=isTRUE(Sys.getenv("ART_GENERATE_HTML") == "TRUE") # Set with Sys.setenv(ART_GENERATE_HTML = "TRUE")

overwrite = FALSE # Use TRUE when creating new tests that need new *_expected data.

test_that("amExample1 from pg 5 in allelematchSuppDoc.pdf is 2.5.3 compatible", code = {

    # Prepare for printing large snapshot files:
    withr::local_options(width=200) # Allow longer lines for the summaries:

    # Follow the instructions from allelematchSuppDoc.pdf:
    # Test the results within curly brackets ("{ ... }") below the instructions.
    data("amExample1")
    expect_snapshot(
        example1 <- amDataset(amExample1, indexColumn="sampleId", ignoreColumn="knownIndividual", missingCode="-99")
    )
    {
        # Ensure that the result is still the same as that from 2.5.3:
        expect_snapshot_value(example1, style="deparse")
    }

    output <- capture.output(
        amUniqueProfile(example1, doPlot=TRUE)
    )
    {
        # Ensure that the result is still the same as that from 2.5.3:
        # cat("\nOutput from amUniqueProfile:\n", output, "\nEnd of output\n", sep="\n    ")

        testthat::expect_match(output, "missing data load for input dataset is 0.005 ", fixed=TRUE, all=FALSE)
        testthat::expect_match(output, "allelic diversity for input dataset is 6.1 ", fixed=TRUE, all=FALSE)
        testthat::expect_match(output, "Best guess for optimal parameter at alleleMismatch=2 OR matchThreshold=0.9 OR cutHeight=0.1", fixed=TRUE, all=FALSE)
        testthat::expect_match(output, "Best guess for optimal parameter at alleleMismatch=2 OR matchThreshold=0.9 OR cutHeight=0.1$", all=FALSE)
        testthat::expect_match(output, "Best guess for unique profile morphology: ZeroSecondMinimum", fixed=TRUE, all=FALSE)
    }

    expect_snapshot(
        uniqueExample1 <- amUnique(example1, alleleMismatch=2, verbose=FALSE), # TODO : Default of verbose changed in 2.5.4
    )
    {
        # testthat::expect_match(output, "assuming genotype columns are in pairs, representing 10 loci$", all=FALSE)

        # Ensure that the result is still the same as that from 2.5.3:
        expect_snapshot_value(uniqueExample1, style="deparse")
    }

    if(HTML) {
        ## Save to disk
        summary.amUnique(uniqueExample1, html=helpHtml("example1_1.html"))
        ## View in default browser
        ## summary.amUnique(uniqueExample1, html=TRUE)
    }

    expect_snapshot(
        summary.amUnique(uniqueExample1, csv=summaryFile <- tempfile("example1_1.csv"))
    )
    {
        # Note that the example from allelematchSuppDoc.pdf, summary(uniqueExample1, csv="example1_1.csv"),
        # doesn't work. summary.amUnique is needed.
        #
        # We add the "summaryFile <- tempfile("example1_1.csv")" to avoid that tests generate unwanted files in the workspace.

        # Ensure that the result is still the same as that from 2.5.3
        expect_snapshot_value(read.csv(summaryFile, colClasses="character"), style="deparse")
        file.remove(summaryFile)
    }

    summary.amUnique(uniqueExample1, csv=summaryFile <- tempfile("example1_2.csv"), uniqueOnly=TRUE)
    {
        # Ensure that the result is still the same as that from 2.5.3
        expect_snapshot_value(read.csv(file=summaryFile, colClasses="character"), style="deparse")
        file.remove(summaryFile)
    }

    example1chk <- amDataset(amExample1, indexColumn="sampleId",
      metaDataColumn="knownIndividual", missingCode="-99")
    {
        # Ensure that the result is still the same as that from 2.5.3:
        expect_snapshot_value(example1chk, style="deparse")
    }


    output = capture.output(
        uniqueExample1chk <- amUnique(example1chk, alleleMismatch=2)
    )
    {
        testthat::expect_match(output, "assuming genotype columns are in pairs, representing 10 loci$", all=FALSE)

        # Ensure that the result is still the same as that from 2.5.3:
        expect_snapshot_value(uniqueExample1chk, style="deparse")
    }

    if(HTML) {
        ## Save to disk
        summary.amUnique(uniqueExample1chk, html=helpHtml("example1_2.html"))
        ## View in default browser
        # summary.amUnique(uniqueExample1chk, html=TRUE)
    }
})

test_that("Loop the Loop", {

  # Here we do positive regression testing to make sure that we still get
  # the same results as previously recorded using testthat::expect_snapshot.
  #
  # The recorded results are stored in a file at tests/testthat/_snap/<name_of_this_file>.md
  #
  # We acquire large test coverage by looping over possible argument values,
  # using one for-loop for each parameter in the function we are testing.
  #
  # TODO for allelematch 2.5.3 :
  #   - There is no validation of doPsib in amUnique
  #   - We get 30 errors "'x' must be atomic"! First occurence: amUnique(amdataExample2, cutHeight=0.9, doPsib=missing, consensusMethod=1)
  #   - Print the unconditional, none-verbose messages using 'message()' rather than 'cat()'?
  #   - cutHeight is changed from 0 to 0.00001 in amUnique. Test separately.


  # Run different data sets with different qualities through the same loops:
  miniExample = data.frame(
    "LOC1a"         = c(11:14),
    "LOC1b"         = c(21:24),
    "LOC2a"         = c(31:33, -99),
    "LOC2b"         = c(41:44)
  )
  data("amExample1")
  data("amExample2") ; amExample2 = amExample2[c(1:20),] # Just keep the first 20 rows to save speed and disk
  data("amExample3") ; amExample3 = amExample3[c(1:20),] # Just keep the first 20 rows to save speed and disk
  data("amExample4") ; amExample4 = amExample4[c(1:20),] # Just keep the first 20 rows to save speed and disk
  data("amExample5") ; amExample5 = amExample5[c(1:20),] # Just keep the first 20 rows to save speed and disk

  amdataMini     = amDataset(miniExample)
  amdataExample1 = amDataset(amExample1, indexColumn="sampleId", metaDataColumn="knownIndividual")
  amdataExample2 = amDataset(amExample2, indexColumn="sampleId", metaDataColumn="knownIndividual")
  amdataExample3 = amDataset(amExample3, indexColumn="sampleId", metaDataColumn="knownIndividual")
  amdataExample4 = amDataset(amExample4, indexColumn="sampleId", metaDataColumn="knownIndividual")
  amdataExample5 = amDataset(amExample5, indexColumn="sampleId", ignoreColumn=c("samplingData", "gender"))


  # We want to run amUnique many times with many combinations of parameters,
  # and we want to compare the results with previous runs. Like this:
  snapshot_amUnique <- function(ds, ...) {

    # Log the call to the snapshot file:
    argstr = helpArgToString(...)
    cmdstr = paste("amUnique(", ds, ", ", argstr, ")", sep="") ; expect_snapshot(cat(cmdstr))

    # Capture any errors reported by allelematch:
    sink(nullfile()) # Block output from 'cat' within allelematch
    ret <- tryCatch(

      # Make the call to allelematch:
      amUnique(amDatasetFocal=get(ds), ...),

      ## If the call fails, return the error message and
      ## the method and arguments that threw the error:
      error = function(e) {
        e_message <- trimws(gsub("[ \t]+", " ", e$message)) # Work around space changes with never versions
        ret = c(paste("\n  Error    : ", e_message,
                      "\n  Test     : ", "test-allelematch_6-amUnique",
                      "\n  Rejected : ", cmdstr, "\n"))

        # Differ between expected and unexpected errors:
        if (!grepl("no clusters formed.\\s+Please set cutHeight lower and run again|'x' must be atomic", e_message, perl=TRUE)) {
          # Some unexpected error happened. Print it to the screen for easier debugging.
          message("\n  ", ret, sep="")

        }

        ret # Return this message from 'tryCatch'
      }
    )
    sink() # Block output from 'cat' within allelematch


    # Log the result to the snapshot file:
    # expect_snapshot_value(ret, style = "json2") # Neither "json2" nor "deparse" works. :-(
    if (class(ret) == "amUnique") {
      expect_snapshot(amCSV.amUnique(ret, csvFile=stdout(), uniqueOnly=FALSE))
    } else {
      expect_snapshot(cat("\n  No amUnique object generated:", ret))
    }

    return(ret)
  }

  # Parameters to amUnique:
  #  amUnique <- function(amDatasetFocal, multilocusMap=NULL, alleleMismatch=NULL, matchThreshold=NULL, cutHeight=NULL, doPsib="missing", consensusMethod=1, verbose=FALSE)
  #
  # We simplify by dropping
  #  - multilocusMap (default is good)
  #  - alleleMismatch and matchTheashold (calculated from cutHeight) and
  #  - verbose
  #
  # Parameters left to loop over:
  #  amUnique <- function(amDatasetFocal, cutHeight=NULL, doPsib="missing", consensusMethod=1)
  #
  # Here comes the loops:
  for (amds in c("amdataMini", "amdataExample1", "amdataExample2", "amdataExample3", "amdataExample4", "amdataExample5")) {
    for (ch in c(0, 0.3, 0.5, 0.7, 0.9, 0.95, 0.99, 1)) { # cutHeight. Relevant Values? Range [0..1]
      for (psi in c("missing", "all")) { # doPsib - Method for calculating Probability of sibling
        for (cons in c(1,2,3,4)) { # consensusMethod
          snapshot_amUnique(amds, cutHeight=ch, doPsib=psi, consensusMethod=cons)
        }
      }
    }
  }
})

test_that("We are running the 3rd edition of testthat", code = {
    # Set by adding the line "Config/testthat/edition: 3" to DESCRIPTION:
    testthat::expect_gte(!!testthat::edition_get(), 3)
})

HTML=isTRUE(Sys.getenv("ART_GENERATE_HTML") == "TRUE") # Set with Sys.setenv(ART_GENERATE_HTML = "TRUE")

overwrite = FALSE # Use TRUE when creating new tests that need new *_expected data.

test_that("amExample3 results from pg 10 in allelematchSuppDoc.pdf are 2.5.3 compatible", code = {

    # Prepare for printing large snapshot files:
    withr::local_options(width=200) # Allow longer lines for the summaries:

    # Follow the instructions from allelematchSuppDoc.pdf, pg 10:
    data("amExample3")
    example3 <- amDataset(amExample3, indexColumn="sampleId",
      metaDataColumn="knownIndividual", missingCode="-99")
    {
        # Ensure that the result is still the same as that from 2.5.3:
        expect_snapshot_value(example3, style="deparse")
    }

    output = capture.output(
        amUniqueProfile(example3, doPlot=TRUE)
    )
    {
        # Ensure that the result is still the same as that from 2.5.3:
        # cat("\nOutput from amUniqueProfile:\n", output, "\nEnd of output\n", sep="\n    ")

        testthat::expect_match(output, "missing data load for input dataset is 0.097 ", fixed=TRUE, all=FALSE)
        testthat::expect_match(output, "allelic diversity for input dataset is 8.2 ", fixed=TRUE, all=FALSE)
        testthat::expect_match(output, "Best guess for optimal parameter at alleleMismatch=6 OR matchThreshold=0.7 OR cutHeight=0.3$", all=FALSE)
        testthat::expect_match(output, "Best guess for unique profile morphology: NonZeroSecondMinimum$", all=FALSE)
        testthat::expect_match(output, "Use extra caution.", fixed=TRUE, all=FALSE)
    }

    output = capture.output(
        uniqueExample3 <- amUnique(example3, alleleMismatch=6)
    )
    {
        testthat::expect_match(output, "assuming genotype columns are in pairs, representing 10 loci$", all=FALSE)

        # Ensure that the result is still the same as that from 2.5.3:
        expect_snapshot_value(uniqueExample3, style="deparse")

        # Generate a summary file:
        summary.amUnique(uniqueExample3, csv=summaryFile <- tempfile("example3_1.csv"))

        # Ensure that the result is still the same as that from 2.5.3
        expect_snapshot_value(read.csv(file=summaryFile, colClasses="character"), style="deparse")
        file.remove(summaryFile)
    }

    if(HTML) {
        summary.amUnique(uniqueExample3, html=helpHtml("example3_1.html"))
    }

    output = capture.output(
        unclassifiedExample3 <- amPairwise(uniqueExample3$unclassified,
                                uniqueExample3$unique, alleleMismatch=7)
    )
    {
        # cat("\n    output='", output, "'\n", sep="")
        # testthat::expect_match(output, "^$", perl=TRUE, all=FALSE) # (Empty output)

        # Ensure that the result is still the same as that from 2.5.3:
        expect_snapshot_value(unclassifiedExample3, style="deparse")

        # Generate a summary file:
        summary.amPairwise(unclassifiedExample3, csv=summaryFile <- tempfile("example3_2.csv"))

        # Ensure that the result is still the same as that from 2.5.3
        expect_snapshot_value(read.csv(file=summaryFile, colClasses="character"), style="deparse")
        file.remove(summaryFile)
    }

    if(HTML) {
        summary.amPairwise(unclassifiedExample3, html=helpHtml("example3_2.html"))
    }


    output = capture.output(
        multipleMatchExample3 <- amPairwise(uniqueExample3$multipleMatch,
                                            uniqueExample3$unique, alleleMismatch=6)
    )
    {
        # testthat::expect_match(output, "^$", all=FALSE) # (Empty output)

        # Ensure that the result is still the same as that from 2.5.3:
        expect_snapshot_value(multipleMatchExample3, style="deparse")

        # Generate a summary file:
        summary.amPairwise(multipleMatchExample3, csv=summaryFile <- tempfile("example3_3.csv"))

        # Re-read the generated .csv file:
        expect_snapshot_value(read.csv(file=summaryFile, colClasses="character"), style="deparse")
        file.remove(summaryFile)
    }

    if(HTML) {
        summary.amPairwise(multipleMatchExample3, html=helpHtml("example3_3.html"))
    }

})

HTML=isTRUE(Sys.getenv("ART_GENERATE_HTML") == "TRUE") # Set with Sys.setenv(ART_GENERATE_HTML = "TRUE")

overwrite = FALSE # Use TRUE when creating new tests that need new *_epected data.

test_that("amExample2 results from pg 8 in allelematchSuppDoc.pdf 2.5.3 compatible", code = {

    # Prepare for printing large snapshot files:
    withr::local_options(width=200) # Allow longer lines for the summaries:

    # Follow the instructions from allelematchSuppDoc.pdf, pg:
    data("amExample2")
    example2 <- amDataset(amExample2, indexColumn="sampleId",
      metaDataColumn="knownIndividual", missingCode="-99")
    {
        # Ensure that the result is still the same as that from 2.5.3:
        expect_snapshot_value(example2, style="deparse")
    }

    output = capture.output(
        amUniqueProfile(example2, doPlot=TRUE)
    )
    {
        # Check that the output to STDOUT is still the same as that from 2.5.3:
        testthat::expect_match(output, "missing data load for input dataset is 0.046 ", fixed=TRUE, all=FALSE)
        testthat::expect_match(output, "allelic diversity for input dataset is 7.9 ", fixed=TRUE, all=FALSE)
        testthat::expect_match(output, "Best guess for optimal parameter at alleleMismatch=3 OR matchThreshold=0.85 OR cutHeight=0.15$", all=FALSE)
        testthat::expect_match(output, "Best guess for unique profile morphology: ZeroSecondMinimum$", all=FALSE)
    }

    output = capture.output(
        uniqueExample2 <- amUnique(example2, alleleMismatch=3)
    )
    {
        # Check that the output to STDOUT is still the same as that from 2.5.3:
        testthat::expect_match(output, "assuming genotype columns are in pairs, representing 10 loci$", all=FALSE)

        # Ensure that the result is still the same as that from 2.5.3:
        expect_snapshot_value(uniqueExample2, style="deparse")

        # Generate a summary file:
        summary.amUnique(uniqueExample2, csv=summaryFile <- tempfile("example2_1.csv"))

        # Ensure that the result is still the same as that from 2.5.3
        expect_snapshot_value(read.csv(file=summaryFile, colClasses="character"), style="deparse")
        file.remove(summaryFile)
    }

    if(HTML) {
        summary.amUnique(uniqueExample2, html=helpHtml("example2_1.html"))
    }

    output = capture.output(
            uniqueExample2 <- amUnique(example2, alleleMismatch=3, doPsib="all")
    )
    {
        testthat::expect_match(output, "assuming genotype columns are in pairs, representing 10 loci$", all=FALSE)

        # Ensure that the result is still the same as that from 2.5.3:
        expect_snapshot_value(uniqueExample2, style="deparse")

        # Generate a summary file:
        summary.amUnique(uniqueExample2, csv=summaryFile <- tempfile("example2_2.csv"))

        # Ensure that the result is still the same as that from 2.5.3
        expect_snapshot_value(read.csv(file=summaryFile, colClasses="character"), style="deparse")
        file.remove(summaryFile)
    }

    if(HTML) {
        summary.amUnique(uniqueExample2, html=helpHtml("example2_2.html"))
    }

    # Copied from test of amExample1:
    example2chk <- amDataset(amExample2, indexColumn="sampleId",
                             metaDataColumn="knownIndividual", missingCode="-99")
    {
        # Ensure that the result is still the same as that from 2.5.3:
        expect_snapshot_value(example2chk, style="deparse")
    }

    output = capture.output(
        uniqueExample2chk <- amUnique(example2chk, alleleMismatch=2)
    )
    {
        testthat::expect_match(output, "assuming genotype columns are in pairs, representing 10 loci$", all=FALSE)

        # Ensure that the result is still the same as that from 2.5.3:
        expect_snapshot_value(uniqueExample2chk, style="deparse")
    }

})

test_that("Loop the Loop", {

  # Here we do positive regression testing to make sure that we still get
  # the same results as previously recorded using testthat::expect_snapshot.
  #
  # The recorded results are stored in a file at tests/testthat/_snap/<name_of_this_file>.md
  #
  # We acquire large test coverage by looping over possible argument values,
  # using one for-loop for each parameter in the function we are testing.
  #
  # TODO for allelematch 2.5.3 :
  #   - Make amUniqueProfile return a class.
  #   - Bug on line 1253: Change "if(length(cutHeight <= 2))" to "if(length(cutHeight) <= 2)"
  #   - Bug on line 1264: Change "if(length(matchThreshold <= 2))" to "if(length(matchThreshold) <= 2)"
  #   - ?amUniqueProfile can be misunderstood to say that one of alleleMismatch, cutHeight and matchThreashold must be present.


  # Run different data sets with different qualities through the same loops:
  miniExample = data.frame(
    "LOC1a"         = c(11:14),
    "LOC1b"         = c(21:24),
    "LOC2a"         = c(31:33, -99),
    "LOC2b"         = c(41:44),
    "LOC3a"         = c(51:53, -99),
    "LOC3b"         = c(61:64)
  )
  data("amExample1")
  data("amExample2") #; amExample2 = amExample2[c(1:20),] # Just keep the first 20 rows to save speed and disk
  data("amExample3") #; amExample3 = amExample3[c(1:20),] # Just keep the first 20 rows to save speed and disk
  data("amExample4") #; amExample4 = amExample4[c(1:20),] # Just keep the first 20 rows to save speed and disk
  data("amExample5") #; amExample5 = amExample5[c(1:20),] # Just keep the first 20 rows to save speed and disk

  amdataMini     = amDataset(miniExample)
  amdataExample1 = amDataset(amExample1, indexColumn="sampleId", metaDataColumn="knownIndividual")
  amdataExample2 = amDataset(amExample2, indexColumn="sampleId", metaDataColumn="knownIndividual")
  amdataExample3 = amDataset(amExample3, indexColumn="sampleId", metaDataColumn="knownIndividual")
  amdataExample4 = amDataset(amExample4, indexColumn="sampleId", metaDataColumn="knownIndividual")
  amdataExample5 = amDataset(amExample5, indexColumn="sampleId", ignoreColumn=c("samplingData", "gender"))


  # We want to run amUniqueProfile many times with many combinations of parameters,
  # and we want to compare the results with previous runs. Like this:
  snapshot_amUniqueProfile <- function(ds, ...) {

    # Log the call to the snapshot file:
    argstr = helpArgToString(...)
    cmdstr = paste("amUniqueProfile(", ds, ", ", argstr, ")", sep="") ; expect_snapshot(cat(cmdstr))

    # Capture any errors reported by allelematch, but drop the output to stdout:
    ret <- tryCatch(

      # Make the call to allelematch:
      capture.output(amUniqueProfile(get(ds), ...)),

      ## If the call fails, return the error message and
      ## describe the method and arguments that threw the error:
      error = function(e) {
        ret = c(paste("\n  Error    : ", e$message, "\n  Rejected : ", cmdstr, "\n"))

        # Differ between expected and unexpected errors:
        if (!grepl("no clusters formed.\\s+Please set cutHeight lower and run again", e$message, perl=TRUE)) {
          # Some unexpected error happened. Print it to the screen for easier debugging.
          message("\n  ", ret, "\n  error class = ", format(class(e)), sep="")

        }

        ret
      }
    )
    ret = paste(ret, collapse = "\n")

    # Log the result to the snapshot file:
    expect_snapshot(cat(ret))

    return(ret)
  }

  # Parameters to amUniqueProfile:
  #  amUniqueProfile <- function(
  #     amDatasetFocal, multilocusMap=NULL, alleleMismatch=NULL, matchThreshold=NULL, cutHeight=NULL,
  #     guessOptimum=TRUE, doPlot=TRUE, consensusMethod=1, verbose=TRUE)
  #
  # We simplify by dropping
  #  - multilocusMap (default is good)
  #  - alleleMismatch (Automatically cykled from [1:(number_of_loci-1)])
  #  - matchThreashold and cutHeight (calculated from alleleMismatch)
  #  - guessOptimum (Keep default=TRUE. Exercises more code)
  #  - doPlot (Keep default=TRUE. Exercises more code)
  #  - consensusMethod (Keep default=1. Assume tested for amCluster)
  #  - verbose (the default, TRUE, just generates noice)
  #
  # Parameters left to loop over:
  #  amUniqueProfile <- function(amDatasetFocal)
  #
  # Here comes the loops:
  withr::local_options(width=200) # Allow longer lines for the summaries:
  for (amds in c("amdataMini", "amdataExample1", "amdataExample2", "amdataExample3", "amdataExample4", "amdataExample5")) {
    snapshot_amUniqueProfile(amds, doPlot=FALSE, verbose=FALSE)
  }

  # One more test to doPlot. Notice that this test will fail spectacularly if the RStudio "Plots" window is too small:
  amdataExample5b = amDataset(amExample5, indexColumn="sampleId", metaDataColumn="samplingData")
  mlMap = c(1, rep(2:11, each=2))
  obj <- amUniqueProfile(
      amdataExample5b,
      multilocusMap=c(1,2,2,3,3,4,4,5,5,6,6,7,7,8,8,9,9,10,10,11,11), # "gender" only has one allele in it's locus
      doPlot=TRUE, verbose=FALSE)
  expect_snapshot(print(obj))

})

test_that("Loop the Loop", {

  # We want to run amCluster many times with many combinations of parameters
  # and we want to compare the results with previous runs. Like this:
  snapshot_amCluster <- function(ds, ...) {

    # Log the call to the snapshot file:
    argstr = helpArgToString(...)
    cmdstr = paste("amCluster(", ds, ", ", argstr, ")", sep="") ; expect_snapshot(cat(cmdstr))

    # Capture any errors reported by allelematch:
    pw <- tryCatch(

      # Make the call to allelematch:
      amCluster(amDatasetFocal=get(ds), ...),

      ## If the call fails, return the error message and
      ## the method and arguments that threw the error:
      error = function(e) {
        e_message <- trimws(gsub("[ \t]+", " ", e$message)) # Work around space changes with never versions
        ret = c(paste("\n  Error    : ", e_message,
                      "\n  Test     : ", "test-allelematch_4-amCluster",
                      "\n  Rejected : ", cmdstr, "\n"))

        # Differ between expected and unexpected errors:
        if (!grepl("no clusters formed.\\s+Please set cutHeight lower and run again", e_message)) {
          # Some unexpected error happened. Print it to the screen for easier debugging.
          message("\n  ", ret, sep="")
        }

        ret # Return this message from 'tryCatch'
      }
    )

    # Log the result to the snapshot file:
    # expect_snapshot(summary.amCluster(pw)) # Too big. :-(
    expect_snapshot_value(pw, style = "json2")

    return(pw)
  }

  # Run different data sets with different qualities through the same loops:
  miniExample = data.frame(
    "LOC1a"         = c(11:14),
    "LOC1b"         = c(21:24),
    "LOC2a"         = c(31:33, -99),
    "LOC2b"         = c(41:44)
  )
  data("amExample1")
  data("amExample2") ; amExample2 = amExample2[c(1:20),] # Just keep the first 20 rows to save speed and disk
  data("amExample3") ; amExample3 = amExample3[c(1:20),] # Just keep the first 20 rows to save speed and disk
  data("amExample4") ; amExample4 = amExample4[c(1:20),] # Just keep the first 20 rows to save speed and disk
  data("amExample5") ; amExample5 = amExample5[c(1:20),] # Just keep the first 20 rows to save speed and disk

  amdataMini     = amDataset(miniExample)
  amdataExample1 = amDataset(amExample1, indexColumn="sampleId", metaDataColumn="knownIndividual")
  amdataExample2 = amDataset(amExample2, indexColumn="sampleId", metaDataColumn="knownIndividual")
  amdataExample3 = amDataset(amExample3, indexColumn="sampleId", metaDataColumn="knownIndividual")
  amdataExample4 = amDataset(amExample4, indexColumn="sampleId", metaDataColumn="knownIndividual")
  amdataExample5 = amDataset(amExample5, indexColumn="sampleId", ignoreColumn=c("samplingData", "gender"))

  # Parameters to loop over:
  #   " amCluster <- function(amDatasetFocal, runUntilSingletons=TRUE, cutHeight=0.3, missingMethod=2, consensusMethod=1, clusterMethod = "complete") {
  # Regarding clusterMethod: "Only 'complete' acceptable. This option remains for experimental reasons". So, we stick with the default.

  # Here comes the loops:
  for (amds in c("amdataMini", "amdataExample1", "amdataExample2", "amdataExample3", "amdataExample4", "amdataExample5")) {
    for (rus in c(TRUE, FALSE)) { # runUntilSingletons
      for (ch in c(0.1, 0.3, 0.5, 0.7, 0.9, 0.95, 0.99)) { # cutHeight. Relevant Values? Range [0..1[ ? NULL?
        for (mis in c(1, 2)) { # missingMethod
          for (cons in c(1,2,3,4)) { # consensusMethod
            snapshot_amCluster(amds, runUntilSingletons=rus, cutHeight=ch, missingMethod=mis, consensusMethod=cons)
          }
        }
      }
    }
  }
})

test_that("Loop the Loop", {
  
  # Here we do positive regression testing to make sure that we still get 
  # the same results as previously recorded using testthat::expect_snapshot.
  #
  # The recorded results are stored in a file at tests/testthat/_snap/<name_of_this_file>.md
  #
  # We acquire large test coverage by looping over possible argument values,
  # using one for-loop for each parameter in the function we are testing.
  #
  # TODO for allelematch 2.5.3 : 
  #   - Make amMatrix return a class.
  #   - Bug on line 1253: Change "if(length(cutHeight <= 2))" to "if(length(cutHeight) <= 2)"
  #   - Bug on line 1264: Change "if(length(matchThreshold <= 2))" to "if(length(matchThreshold) <= 2)"
  #   - ?amMatrix can be misunderstood to say that one of alleleMismatch, cutHeight and matchThreashold must be present.
  
  
  # Run different data sets with different qualities through the same loops:
  miniExample = data.frame(
    "LOC1a"         = c(11:14),
    "LOC1b"         = c(21:24),
    "LOC2a"         = c(31:33, -99),
    "LOC2b"         = c(41:44),
    "LOC3a"         = c(51:53, -99),
    "LOC3b"         = c(61:64)
  )
  data("amExample1")
  data("amExample2") ; amExample2 = amExample2[c(1:20),] # Just keep the first 20 rows to save speed and disk
  data("amExample3") ; amExample3 = amExample3[c(1:20),] # Just keep the first 20 rows to save speed and disk
  data("amExample4") ; amExample4 = amExample4[c(1:20),] # Just keep the first 20 rows to save speed and disk
  data("amExample5") ; amExample5 = amExample5[c(1:20),] # Just keep the first 20 rows to save speed and disk
  
  amdataMini     = amDataset(miniExample)
  amdataExample1 = amDataset(amExample1, indexColumn="sampleId", metaDataColumn="knownIndividual")
  amdataExample2 = amDataset(amExample2, indexColumn="sampleId", metaDataColumn="knownIndividual")
  amdataExample3 = amDataset(amExample3, indexColumn="sampleId", metaDataColumn="knownIndividual")
  amdataExample4 = amDataset(amExample4, indexColumn="sampleId", metaDataColumn="knownIndividual")
  amdataExample5 = amDataset(amExample5, indexColumn="sampleId", ignoreColumn=c("samplingData", "gender"))
  
  
  # We want to run amMatrix many times with many combinations of parameters,
  # and we want to compare the results with previous runs. Like this:
  snapshot_amMatrix <- function(ds, ...) {
    
    # Log the call to the snapshot file:
    argstr = helpArgToString(...)
    cmdstr = paste("amMatrix(", ds, ", ", argstr, ")", sep="") ; expect_snapshot(cat(cmdstr))
    
    # Capture any errors reported by allelematch:
    ret <- tryCatch( 
      
      # Make the call to allelematch:  
      capture.output(amMatrix(get(ds), ...)),
      
      ## If the call fails, return the error message and
      ## describe the method and arguments that threw the error:
      error = function(e) {
        ret = c(paste("\n  Error    : ", e$message, "\n  Rejected : ", cmdstr, "\n"))
        
        # Differ between expected and unexpected errors:
        if (!grepl("SOME TO BE DEFINED KNOWN ERROR", e$message, perl=TRUE)) {
          # Some unexpected error happened. Print it to the screen for easier debugging.
          message("\n  ", ret, sep="")
        }
        
        ret
      }
    )
    ret = paste(ret, collapse = "\n")
    
    # Log the result to the snapshot file:
    expect_snapshot(cat(ret))
    
    return(ret)
  }
  
  # Parameters to amMatrix:
  #   amMatrix <- function(amDatasetFocal, missingMethod=2)
  #
  # Here comes the loops:
  for (amds in c("amdataMini", "amdataExample1", "amdataExample2", "amdataExample3", "amdataExample4", "amdataExample5")) {
    for (mm in c(1,2)) {
      snapshot_amMatrix(amds, missingMethod=mm)
    }
  }
})


test_that("Value from amMatrix() remains stable", {
  
  # data(amExample1)
  # expect_equal(!!dim(amExample1), c(20,22))
  # 
  # amdata = amDataset1_1= amDataset(amExample1, indexColumn = "sampleId", metaDataColumn = "knownIndividual")
  # expect_snapshot_value(amdata, style = "json2")
  
  # Create valid miniature input sample:    
  sample = miniExample = data.frame(
    "LOC1a"         = c(11:14),
    "LOC1b"         = c(21:24),
    "LOC2a"         = c(31:33, -99),
    "LOC2b"         = c(41:44)
  )
  
  # Remember how the matrix is calculated for regression testing:
  {
    mx = amMatrix1 = amMatrix(amDataset(sample))
    expect_snapshot_value(mx, style = "json2")
    
    expect_snapshot(
      list(
        mx1 = amMatrix(amDataset(sample), missingMethod = 1),
        mx2 = amMatrix(amDataset(sample), missingMethod = 2)
      )
    )
  }
  
  
  # Detect changes in the calculation of an amMatrix
  # whilst regression testing:
  data("amExample2") # Good quality data set
  amdataExample2 <- amDataset(amExample2, indexColumn="sampleId",
                              metaDataColumn="knownIndividual", missingCode="-99")
  {
    expect_snapshot(mx21 <- amMatrix(amdataExample2, 1))
    expect_snapshot_value(mx21, style = "json2")
    
    expect_snapshot(mx22 <- amMatrix(amdataExample2, 2))
    expect_snapshot_value(mx22, style = "json2")
  }
  
  ## The following amMatrix:es get too big for expect_snapshot_value :-( 
  # data("amExample3") # Marginal quality data set
  # amdataExample3 <- amDataset(amExample3, indexColumn="sampleId",
  #                             metaDataColumn="knownIndividual", missingCode="-99")
  # {
  #   expect_snapshot(mx31 <- amMatrix(amdataExample3, missingMethod = 1))
  #   expect_snapshot_value(mx31, style = "json2")
  # }
  # 
  # data("amExample4") # Poor quality example
  # amdataExample4 <- amDataset(amExample4, indexColumn="sampleId",
  #                             metaDataColumn="knownIndividual", missingCode="-99")
  # {
  #   expect_snapshot(mx41 <- amMatrix(amdataExample4, 1))
  #   expect_snapshot_value(mx41, style = "json2")
  # }
  # 
  # data("amExample5") # Wildlife example
  # amdataExample5 <- amDataset(amExample5, indexColumn="sampleId",
  #                             metaDataColumn="samplingData", missingCode="-99")
  # {
  #   expect_snapshot(mx51 <- amMatrix(amdataExample5, missingMethod = 1))
  #   expect_snapshot_value(mx51, style = "json2")
  # }
  
})

# Run tests on [ggSample], a large real life data set.
#
# The tests are quite time consuming (approx one minute per test).
# They can therefore be skipped by setting "ART_SKIP_SLOW" to "TRUE"
# from e.g. the RStudio Console:
#
#       Type Sys.setenv(ART_SKIP_SLOW = "TRUE") to skip
#
#       Type Sys.unsetenv("ART_SKIP_SLOW") to enable again
#
test_that("amUnique(matchThreshold=0.9) for dataset ggSample", code = {

    skip_if(Sys.getenv("ART_SKIP_SLOW") == "TRUE")
    # Type Sys.setenv(ART_SKIP_SLOW = "TRUE") from the console to skip,
    # Type Sys.unsetenv("ART_SKIP_SLOW") to enable again

    # Prepare for printing very large snapshot files:
    withr::local_options(width=3000) # Allow longer lines for the summaries:

    # Load the large gg-style sample file:
    data("ggSample")
    expect_identical(dim(ggSample), c(1658L, 187L))

    # Load the sample into an allelematch amDataset:
    expect_snapshot(ggDataset <- allelematch::amDataset(ggSample, indexColumn=1, missingCode="-99"))

    # We want to run amUnique many times with many combinations of parameters,
    # and we want to compare the results with previous runs. Like this:
    snapshot_amUnique <- function(ds, ...) {

        # Log the call to the snapshot file:
        argstr = helpArgToString(...)
        cmdstr = paste("obj <- amUnique(", ds, ", ", argstr, ")", sep="") ; expect_snapshot(cat(cmdstr))

        # Make the call to allelematch:
        sink(nullfile()) # Block output from 'cat' within allelematch
        obj = amUnique(amDatasetFocal=get(ds), ...)
        sink()

        # amCSV.amUnique should have the same output as before
        tmp = tempfile(paste(ds, "_", sep=""), fileext=".csv")
        expect_snapshot(summary.amUnique(obj, csv=tmp))
        expect_snapshot(format(read.csv(tmp)))
        file.remove(tmp)

        # # amHTML.amUnique should have the same output as before
        # tmp = tempfile(paste(ds, "_", sep=""), fileext=".html")
        # expect_snapshot(amHTML.amUnique( obj, htmlFile=tmp))
        # expect_snapshot(
        #     cat(
        #         sub("summary generated: </b><em>.+?</em>",
        #             "summary generated: </b><em>(date)</em>",
        #             gsub("(\\t| )+?(\\n|$)","\\2",
        #                  readLines(tmp, warn=FALSE),
        #                  perl=TRUE),
        #             perl=TRUE),
        #         sep="\n")
        #
        # )
        # file.remove(tmp)


        invisible(NULL)
    }

    snapshot_amUnique("ggDataset", matchThreshold=0.9)
    snapshot_amUnique("ggDataset", alleleMismatch=15)

})

#' @keywords internal
"_PACKAGE"

## usethis namespace: start
#    List imports. Entered into the "Imports:" section of DESCRIPTION file by roxygen2.
#    Order: "testthat depends on withr, so withr must come first"
#' @import utils
#' @import digest
#' @importFrom R6 R6Class
#' @import remotes
#' @import withr
#' @import testthat
#' @import allelematch
## usethis namespace: end
NULL


#' @name amregtest
#'
#' @title Package Overview
#'
#' @description
#' Package 'amregtest' automates regression testing of package [allelematch][allelematch::allelematch-package].
#'
#' The API is simple. There are only three functions:
#' \tabular{clcl}{
#'  `  ` \tab \code{\link{artRun}()}`  `\tab `  ` \tab Executes the test, or a subset of the tests\cr
#'  `  ` \tab \code{\link{artList}()}\tab \tab Lists the available tests without running them\cr
#'  `  ` \tab \code{\link{artVersion}()}\tab \tab Shows the used versions of 'allelematch' and 'amregtest'\cr
#' }
#'
#' The prefix "art" is short for "Allelematch Regression Test".
#'
#' See \link{artData} for a description of data sets used as input.

#'
#' @references \url{https://github.com/cran/allelematch}
#' @references \href{https://github.com/cran/allelematch/blob/2.5.1/inst/doc/allelematchSuppDoc.pdf}{allelematchSuppDoc.pdf}
NULL

test_that("We are running the 3rd edition of testthat", code = {
  # If this tests fails, then call
  #   usethis::use_testthat(3)
  # to configure DESCRIPTION to use 3rd edition of 'testthat'.
  testthat::expect_gte(!!testthat::edition_get(), 3)
})

test_that("amExamples have not changed in allelematch", {

    # Calculate a checksum for data stored under ./data/ in a package:
    env <- environment() # Use a local environment to avoid polluting the global environment with data sets
    md5sum <- function(name, package) {
        stopifnot(is.character(name) && is.character(package))
        utils::data(list = c(name), package=package, envir = env)[1]
        cs = digest::digest(get(name, envir = env))
        return(cs)
    }

    # 2.6.0 of allelematch introduced some cosmetic changes,
    # including changing the column name "gender" to "sex" in amExample5.
    # This changed the md5sum for that data set, so we need to check the version
    # of allelematch to know which md5sum to expect.
    amversion <- packageVersion("allelematch")

    expect_identical(md5sum("amExample1", package="allelematch"), '25108ea88af5cc916ed887c82eb89840')
    expect_identical(md5sum("amExample2", package="allelematch"), 'a438a316c63bbc024c3fe0eb564c4edb')
    expect_identical(md5sum("amExample3", package="allelematch"), '242ef242fc6afd413d13f7f7739823af')
    expect_identical(md5sum("amExample4", package="allelematch"), 'd7a34f4319c15e8042fe01cdfed18bc3')
    if(amversion < "2.6.0") { # Versions 2.5.5 and earlier had a different md5sum for amExample5
      expect_identical(md5sum("amExample5", package="allelematch"), 'cce57ddcaaa4ae31902b6036a0a90f8e')
    } else {
      # Only difference is the change in one column name from "gender" to "sex":
      expect_identical(md5sum("amExample5", package="allelematch"), '5f481f6287de5d5cc05277b645ba0642')
    }

    # Load the 2.5.5 version for use in our tests where we assume "gender":
    data("amExample5", package="amregtest", envir = env) # Load the 2.5.5 version from amregtest

    expect_identical(dim(amExample1), c( 20L, 22L))
    expect_identical(dim(amExample2), c(148L, 22L))
    expect_identical(dim(amExample3), c(319L, 22L))
    expect_identical(dim(amExample4), c(307L, 22L))
    expect_identical(dim(amExample5), c(335L, 23L))
})

test_that("See how an object of class amDataset is built:", {

  # Miniature input sample:
  miniExample = data.frame(
    sampleId        = c(1:4),
    knownIndividual = c("A","A","B","  C"),
    "dismiss "      = c("Rain", " drops", " keep", " fallin' on my head "),
    "LOC1a"         = c(11:14),
    "LOC1b"         = c(21:24),
    "LOC2a"         = c(31:33, -88),
    "LOC2b"         = c(41:44)
  )
  {
    example = miniExample
    expect_identical(!!dim(example), c(4L,7L)) # Four rows, 7 columns
    # Column names:
    # Note that space (" ") has been changed to period (".") after dismiss:
    expect_identical(!!names(example),     c("sampleId","knownIndividual","dismiss.","LOC1a","LOC1b","LOC2a","LOC2b"))
    expect_identical(!!names(example[1,]), c("sampleId","knownIndividual","dismiss.","LOC1a","LOC1b","LOC2a","LOC2b"))
    # Column content (sampleId):
    expect_identical(!!example$sampleId, c(1L, 2L, 3L, 4L))
    expect_identical(!!example[, 1],     c(1L, 2L, 3L, 4L))
    # Row content:
    expect_identical(!!as.vector(as.character(example[1, ])), c(1,"A","Rain","11","21","31","41"))
    expect_identical(!!as.vector(example[1, ]),
                     list(sampleId = 1L, knownIndividual = "A", "dismiss." = "Rain",
                          LOC1a = 11L, LOC1b = 21L, LOC2a = 31, LOC2b = 41L))
    expect_identical(!!example[1, ],
                     as.data.frame(list(sampleId = 1L, knownIndividual = "A", "dismiss." = "Rain",
                          LOC1a = 11L, LOC1b = 21L, LOC2a = 31, LOC2b = 41L)))

    expect_snapshot(print(miniExample))
  }

  # Make amDataset with all optional parameters defaulted:
  expect_snapshot(miniDataset1 <- amDataset(miniExample))
  expect_snapshot(print.amDataset(miniDataset1))
  {
    amDataset = miniDataset1
    expect_identical(!!class(amDataset), "amDataset")
    expect_identical(!!names(amDataset), c("index", "multilocus", "missingCode"))
    expect_identical(!!dim(amDataset$multilocus), c(4L,7L)) # Same as miniExample
    expect_identical(!!is.matrix(amDataset$multilocus), TRUE)

    # Column names stay the same:
    expect_identical(!!names(amDataset[["multilocus"]][1, ]), c("sampleId","knownIndividual","dismiss.", # NB: "knownIndividual" and "dismiss" columns still left
                                                            "LOC1a","LOC1b","LOC2a","LOC2b"))

    # Column contents:
    # Note that the spaces have been stripped for columns expected to be multilocus data:
    expect_identical(!!amDataset[["multilocus"]][, "dismiss."], c("Rain","drops","keep","fallin'onmyhead")) # NB: spaces stripped

    # Data columns are converted from integer to character:
    expect_identical(!!amDataset[["multilocus"]][, "LOC1a"],    c("11", "12", "13", "14"))

    # Defaulted data:
    expect_identical(!!amDataset$index, c("AAA","AAB","AAC", "AAD")) # Auto-generated individual IDs
    expect_identical(!!amDataset$metaData, NULL)
    expect_identical(!!amDataset$missingCode, "-99")
    expect_identical(!!amDataset$metadataColumn, NULL)

    expect_snapshot_value(amDataset, style = "json2") # Stored under ./tests/testthat/_snaps/allelematch.md
  }

  # Make amDataset with all optional parameters set:
  expect_snapshot(miniDataset2 <- amDataset(miniExample, missingCode="-88", indexColumn="sampleId", metaDataColumn="knownIndividual", ignoreColumn="dismiss."))
  expect_snapshot(print.amDataset(miniDataset2))
  {
    amDataset = miniDataset2
    expect_identical(!!class(amDataset), "amDataset")
    expect_identical(!!names(amDataset), c("index", "metaData", "multilocus", "missingCode"))
    expect_identical(!!dim(amDataset$multilocus), c(4L,4L)) # Three columns from miniExample dropped

    # Check that "knownIndividual" and "dismiss" columns have been dropped:
    expect_identical(!!is.matrix(amDataset$multilocus),       TRUE)
    expect_identical(!!rownames(amDataset$multilocus),        NULL)
    expect_identical(!!colnames(amDataset$multilocus),        c("LOC1a","LOC1b","LOC2a","LOC2b"))
    expect_identical(!!names(amDataset[["multilocus"]][1, ]), c("LOC1a","LOC1b","LOC2a","LOC2b"))
    expect_identical(!!amDataset[["multilocus"]][, "LOC2b"],  c("41", "42", "43", "44")) # NB: Integers converted to character
    expect_identical(!!amDataset$multilocus[, "LOC2b"],       c("41", "42", "43", "44")) # NB: Integers converted to character
    expect_identical(!!amDataset$index, c("1","2","3","4")) # NB: Integers converted to character
    expect_identical(!!amDataset$metaData, c("A","A","B","  C")) # Spaces not stripped from metadata
    expect_identical(!!amDataset$missingCode, "-88")

    expect_snapshot_value(amDataset, style = "json2")
  }

  # Make amDataset with all column parameters set as integers rather than characters:
  expect_snapshot(miniDataset3 <- amDataset(miniExample, missingCode="-88", indexColumn=1, metaDataColumn=2, ignoreColumn=3))
  expect_snapshot(print.amDataset(miniDataset2))
  {
    # TODO : Catch none-character values for missingCode!!
    amDataset = miniDataset3
    expect_identical(amDataset, miniDataset2)

    expect_snapshot_value(amDataset, style = "json2")
  }
})


test_that("Different data types for arg to 'missingCode' give same result", {

  # Create valid miniature input sample that contains both NA and "NA":
  sample = miniExample = data.frame(
    "LOC1a"         = c(11:14),
    "LOC1b"         = c(21:23, NA),             # NA integer
    "LOC2a"         = c("31", "32", "33",  NA), # NA character
    "LOC2b"         = c("41", "42", "43", "NA") # string constant "NA"
  )

  # Make sure we still keep the difference between NA and "NA",
  # even though it isn't visible with 'print()':
  {
    expect_true( is.na(sample$LOC1b[4])) # NA integer
    expect_true( is.na(sample$LOC2a[4])) # NA character
    expect_false(is.na(sample$LOC2b[4])) # String constant "NA"
    expect_identical(sum(is.na(sample)), 2L)

    expect_type( sample$LOC1b[4], "integer")
    expect_type( sample$LOC2a[4], "character")
    expect_type( sample$LOC2b[4], "character")

    expect_identical(sample$LOC2b[4], c("NA"))
  }

  # Make sure that both missingCode and data are stored as character:
  expect_snapshot(ds1 <- amDataset(sample, missingCode = "NA"))
  {
    ds = ds1
    expect_identical(ds$missingCode, "NA")
    expect_type(ds$multilocus, "character")
    expect_identical(sum(unlist(ds$multilocus) == "NA"), 3L) # All 3 NA values now as strings
    expect_snapshot_value(ds, style = "deparse") # style "json2" de-serializes "NA" to NA
  }

  # Make sure arg missingCode = NA is converted to $missingCode="NA"
  expect_snapshot(ds2 <- amDataset(sample, missingCode = NA))
  {
    ds = ds2
    expect_snapshot_value(ds, style = "deparse")  # style "json2" de-serializes "NA" to NA
    expect_identical(ds$missingCode, "NA")
    expect_type(ds$multilocus, "character")
    expect_identical(sum(is.na(ds$multilocus)), 0L)
    expect_identical(sum(unlist(ds$multilocus) == "NA"), 3L) # All 3 NA values now as strings
    expect_identical(sum(is.na(ds)), 0L) # Not sure what this does, exactly
  }

  expect_identical(ds1, ds2)
})

# Shadow test_that to detect TEMP directory leaks per test.
#
# testthat helper files are sourced into an environment that all test files
# inherit, so this definition shadows testthat::test_that for every test
# without requiring any changes to individual test files.
#
# For each test it:
#   1. Snapshots tempdir() contents before the test runs.
#   2. Delegates to the real testthat::test_that.
#   3. On exit, records any new files found in tempdir() for deferred cleanup.
#
# After all tests have run, a single cleanup step fires — see setup-temp-leaks.R,
# which registers the deferred action once the test session is initialised.

# Accumulates leaked paths across all tests.
.art_leaked <- character(0)

test_that <- function(desc, code) {
  before <- list.files(tempdir(), recursive = TRUE, full.names = TRUE,
                       all.files = TRUE)

  on.exit({
    after  <- list.files(tempdir(), recursive = TRUE, full.names = TRUE,
                         all.files = TRUE)
    leaked <- setdiff(after, before)

    # On non-Windows, allelematch's amHTML.* functions build paths with a
    # Windows-style "\" separator, e.g. paste(tempdir(), "\\", "amXXX.htm").
    # Unix path parsing splits only on "/", so the file lands in
    # dirname(tempdir()) with a name like "RtmpXXX\amXXX.htm" — one level
    # above tempdir() and invisible to the scan above.  Catch those here.
    if (.Platform$OS.type != "windows") {
      escaped_htm <- list.files(
        dirname(tempdir()),
        pattern = paste0("^", basename(tempdir()), "\\\\am.*\\.htm$"),
        full.names = TRUE
      )
      leaked <- c(leaked, escaped_htm)
    }
    leaked <- leaked[file.exists(leaked)]   # race-condition guard
    keep_files <- toupper(Sys.getenv("ART_KEEP_LEAKED_FILES_IN_TEMP")) == "TRUE"

    if (length(leaked) > 0) {
      if (!keep_files) {
        message(sprintf(
          "Cleaning up %d leaked TEMP file(s):\n  %s",
          length(leaked),
          paste(leaked, collapse = "\n  ")
        ))
        graphics.off()
        suppressWarnings(file.remove(leaked))
      } else if (length(leaked) > 0 && keep_files) {
        # Accumulate for deferred removal — do NOT remove here so that
        # HTML files etc. remain openable in a browser until the suite ends.
        .art_leaked <<- c(.art_leaked, leaked)  # Accumulate for deferred removal

        message(sprintf(
          "Keeping %d leaked TEMP file(s) for inspection:\n  %s",
          length(leaked),
          paste(leaked, collapse = "\n  ")
        ))
      }
    }
  }, add = TRUE)

  # Forward the unevaluated code block to the real test_that.
  # bquote(.(x)) splices the captured expression inline so that
  # testthat::test_that's own substitute() sees the original test code.
  eval(
    bquote(testthat::test_that(.(desc), .(substitute(code)))),
    envir = parent.frame()
  )
}

test_that("Validation of arguments to amPairwise() is working", {

  # Create valid miniature input sample:
  miniExample1 = data.frame(
    "LOC1a"         = c(11:14),
    "LOC1b"         = c(21:24),
    "LOC2a"         = c(31:33, -99),
    "LOC2b"         = c(41:44)
  )
  {
    expect_snapshot(print(miniExample1))
    expect_snapshot(amdataMini1 <- amDataset(miniExample1))
    expect_snapshot(print.amDataset(amdataMini1))
    expect_snapshot_value(amdataMini1, style="json2")

    expect_identical(class(amdataMini1), "amDataset")
    expect_true(inherits(amdataMini1, "amDataset"))
  }

  # Create a miniature input sample with extra, odd column:
  oddExample2 = data.frame(
    "LOC1a"         = c(11:14),
    "LOC1b"         = c(21:24),
    "LOC2a"         = c(31:33, -99),
    "LOC2b"         = c(41:44),
    "LOC3a"         = c(51:53, -99)  # Extra, odd column
  )
  {
    expect_snapshot(print(oddExample2))
    expect_snapshot(amdataOdd2 <- amDataset(oddExample2))
    expect_snapshot(print.amDataset(amdataOdd2))
    expect_snapshot_value(amdataOdd2, style = "json2")
  }

    # Parameters are:
  # (amDatasetFocal, amDatasetComparison=amDatasetFocal, alleleMismatch=NULL, matchThreshold=NULL, missingMethod=2) {

  # Test checks against invalid first param to amDataSet(multilocusDataset):
  {
    expect_error(amPairwise(NULL),  "allelematch:\\s+amDatasetFocal and amDatasetComparison must be an object of class \"amDataset\"")
    expect_error(amPairwise(NA),    "allelematch:\\s+amDatasetFocal and amDatasetComparison must be an object of class \"amDataset\"")
    expect_error(amPairwise(2),     "allelematch:\\s+amDatasetFocal and amDatasetComparison must be an object of class \"amDataset\"")
    expect_error(amPairwise(hipphopp=2), "unused argument")
    expect_error(amPairwise(amdataOdd2), "allelematch:\\s+please specify alleleMismatch OR matchThreshold.")

    # An odd number of columns does not generate an error. Maybe it should?
    # expect_error(amPairwise(amdataOdd2, alleleMismatch=5), "allelematch:\\s+there are an odd number of genotype columns in amDatasetFocal")
  }

  # Test blocking of invalid arguments for second parameter, amDatasetComparison:
  amdata = amdataMini1
  {
    must_be_amDataset = 'allelematch:\\s+amDatasetFocal and amDatasetComparison must be an object of class "amDataset"'

    expect_error(amPairwise(amdata, amDatasetComparison = 0),   must_be_amDataset)
    expect_error(amPairwise(amdata, amDatasetComparison = 3),   must_be_amDataset)
    expect_error(amPairwise(amdata, amDatasetComparison = NULL),must_be_amDataset)
    expect_error(amPairwise(amdata, amDatasetComparison = NA),  must_be_amDataset)
    expect_error(amPairwise(amdata, amDatasetComparison = ""),  must_be_amDataset)
    expect_error(amPairwise(amdata, 1.00000000000000001),       must_be_amDataset)
    expect_error(amPairwise(amdata, 1,0),                       must_be_amDataset)
    expect_error(amPairwise(amdata, amDatasetComparison = amdataOdd2),  "allelematch:\\s+please specify alleleMismatch OR matchThreshold")
    expect_error(amPairwise(amdata, amDatasetComparison = amdataOdd2, alleleMismatch=2),  "amDatasetFocal and amDatasetComparison must have the same number of columns / loci")
    expect_error(amPairwise(amdata, amDatasetComparison = amdataOdd2, matchThreshold=0.5),"amDatasetFocal and amDatasetComparison must have the same number of columns / loci")
  }

  # Use the valid amdata to test blocking of other invalid arguments:
  {
    expect_error(amPairwise(amdata, missingMethod = 0),   "allelematch:\\s+missingMethod must equal 1 or 2")
    expect_error(amPairwise(amdata, missingMethod = 3),   "allelematch:\\s+missingMethod must equal 1 or 2")
    expect_error(amPairwise(amdata, missingMethod = NULL),"argument is of length zero")
    expect_error(amPairwise(amdata, missingMethod = NA),  "allelematch:\\s+missingMethod must equal 1 or 2")
    expect_error(amPairwise(amdata, missingMethod = "2.0"), "allelematch:\\s+missingMethod must equal 1 or 2")
    expect_error(amPairwise(amdata, missingMethod = " 2"), "allelematch:\\s+missingMethod must equal 1 or 2")
    expect_error(amPairwise(amdata, missingMethod = "2 "), "allelematch:\\s+missingMethod must equal 1 or 2")
    expect_error(amPairwise(amdata, missingMethod = 2,0),  must_be_amDataset)
    expect_error(amPairwise(amdata, missingMethod = 2.000000000000001), "allelematch:\\s+missingMethod must equal 1 or 2")

    expect_snapshot_value(amdata, style = "json2")

    # The following calls did not throw an error. Maybe they should?
    # expect_error(amPairwise(amdata, all=2), "unused argument")   # 'all' matches 'alleleMismatch'

  }

  # The following arguments to missingMethod pass. Maybe they shouldn't:
  expect_identical(class(amdata), "amDataset")
  expect_snapshot(pw0 <- amPairwise(amdata, alleleMismatch=2))
  # expect_snapshot(pw0 <- amPairwise(amdataMini1, matchThreshold=0,5) )    # Expected datatype
  {

    # Make sure that different argument types for the parameter "missingMethod"
    # doesn't change the calculation of the amPairwise:
    # expect_identical(pw0, amPairwise(amdata, matchThreshold=0.5, 1.0))  # Unexp
    expect_identical(pw0, amPairwise(amdata, matchThreshold=0.5))

    expect_snapshot_value(pw0, style = "json2")
  }
})

#
# This file is called 'tdd-xxxx.R' rather than 'test-xxx.R'.
# It is therefore not included the normal package tests.
# But they can still be executed manually.
#
# The content of a 'tdd-xxxx.R' contains tests that currently fail because of
# some flaws in allelematch.
#
# The intention is to fix the flaws by using the Test Driven Design (TDD) approach.
#
# In TDD, you (1) make sure the flaw can be reproduced and detected with a new test.
# You then (2) fix the flaw and (3) make sure the test are now working.
# You  then (4) move the tests to 'test-xxx.R' to ensure the fixed problems
# never occur again.
#
test_that("Test Driven Design, TDD - Tests that are still failing", {

  # Create valid miniature input sample:
  miniExample1 = data.frame(
    "LOC1a"         = c(11:14),
    "LOC1b"         = c(21:24),
    "LOC2a"         = c(31:33, -99),
    "LOC2b"         = c(41:44)
  )
  {
    amdata <- amDataset(miniExample1)
  }

  # These bad parameters should have caused errors:
  {
    # These bad parameters should have caused errors:

    # Correct argument value, but wrong data type:
    expect_identical(class(amdata), "amDataset")

    # TODO: These arguments do not throw errors. But maybe they should:
    expect_error(amPairwise(amdata, missingMethod = "2", alleleMismatch=2), "missingMethod must equal 1 or 2")
    expect_error(amPairwise(amdata, missingMethod = 2.0, alleleMismatch=2), "missingMethod must equal 1 or 2")

    # This call does not generate any error:
    amPairwise(amdata, matchThreshold=0.5)

    # TODO: The following error messages are invalid:
    must_be_amDataset = 'amDatasetFocal and amDatasetComparison must be an object of class "amDataset"'
    expect_error(amPairwise(amdata, matchThreshold=0,5), must_be_amDataset) # /NOK
    expect_error(amPairwise(amdata, matchThreshold="0,5"), "matchThreshold`: non-numeric argument to binary operator") # /OK
    expect_error(amPairwise(amdata, amDatasetComparison=amdata, matchThreshold=0,5), "please specify alleleMismatch OR matchThreshold.") # /NOK
    expect_error(amPairwise(amDatasetFocal=amdata, amDatasetComparison=amdata, matchThreshold=0,5), "please specify alleleMismatch OR matchThreshold.") # /NOK
    expect_error(amPairwise(amdata, missingMethod = 1.00000000000000001, matchThreshold=0,5), must_be_amDataset) # /NOK
    expect_error(amPairwise(amdata, matchThreshold=0,5, "1"), must_be_amDataset) # /NOK
    expect_error(amPairwise(amdata, matchThreshold=0,5, 1.0), must_be_amDataset) # /NOK
    expect_error(amPairwise(amdata, matchThreshold=0,5, 1.0000000000000001), must_be_amDataset) # /NOK

  }

  # The following arguments to missingMethod pass. Maybe they shouldn't:
  expect_identical(class(amdata), "amDataset")
  expect_snapshot(pw0 <- amPairwise(amdata, alleleMismatch=2))
  # expect_snapshot(pw0 <- amPairwise(amdataMini1, matchThreshold=0,5) )    # Expected datatype
  {

    # Make sure that different argument types for the parameter "missingMethod"
    # doesn't change the calculation of the amPairwise:
    # expect_identical(pw0, amPairwise(amdata, matchThreshold=0.5, 1.0))  # Unexp
    expect_identical(pw0, amPairwise(amdata, matchThreshold=0.5))

    expect_snapshot_value(pw0, style = "json2")
  }
})

#
# This file is called 'tdd-xxxx.R' rather than 'test-xxx.R'.
# It is therefore not included the normal package tests.
# But they can still be executed manually.
#
# The content of a 'tdd-xxxx.R' contains tests that currently fail because of
# some flaws in allelematch.
#
# The intention is to fix the flaws by using the Test Driven Design (TDD) approach.
#
# In TDD, you (1) make sure the flaw can be reproduced and detected with a new test.
# You then (2) fix the flaw and (3) make sure the test are now working.
# You  then (4) move the tests to 'test-xxx.R' to ensure the fixed problems
# never occur again.
#
test_that("Test Driven Design, TDD - Tests that are still failing", {

  # Create valid miniature input sample:
  miniExample = data.frame(
    # sampleId        = c(1:4),
    # knownIndividual = c("A","A","B","  B"),
    # dismiss         = c("Rain", " drops", " keep", " fallin' on my head "),
    "LOC1a"         = c(11:14),
    "LOC1b"         = c(21:24),
    "LOC2a"         = c(31:33, -99),
    "LOC2b"         = c(41:44)
  )
  amdata <- amDataset(miniExample)

  # These parameters should probably have caused errors:
  {
    # Correct argument value, but wrong data type:
    expect_error(amMatrix(amdata, missingMethod = "2"), "missingMethod must equal 1 or 2")
    expect_error(amMatrix(amdata, missingMethod = 2.0), "missingMethod must equal 1 or 2")
    expect_error(amMatrix(amdata, missingMethod = 1.00000000000000001), "missingMethod must equal 1 or 2")
    expect_error(amMatrix(amdata, 1.00000000000000001), "missingMethod must equal 1 or 2")
  }

  # The following arguments to missingMethod pas. Maybe they shouldn't:
  expect_snapshot(mx0 <- amMatrix(amdataMini, 1) )    # Expected datatype
  {
    # Make sure that different argument types for the parameter "missingMethod"
    # doesn't change the calculation of the amMatrix:
    expect_identical(mx0, amMatrix(amdata, "1"))  # Unexp
    expect_identical(mx0, amMatrix(amdata, 1.0))  # Unexp
    expect_identical(mx0, amMatrix(amdata, 1.0000000000000001))  # Unexp

    expect_snapshot_value(mx0, style = "json2")
  }
})

test_that("Print", {

  # Exercise the three ways to print the output from 'amUnique()' with
  # input data from different amDatasets:

  # Set up the different amDataSets:
  miniExample = data.frame(
    "LOC1a"         = c(11:14),
    "LOC1b"         = c(21:24),
    "LOC2a"         = c(31:33, -99),
    "LOC2b"         = c(41:44)
  )
  data("amExample5") ; amExample5 = amExample5[c(1:20),] # Just keep the first 20 rows to save speed and disk

  sink(nullfile()) # Drop cat output to stdout
  objMini     = amUnique(amDataset(miniExample), matchThreshold=0.5)
  objExample5 = amUnique(amDataset(amExample5, indexColumn="sampleId", ignoreColumn=c("samplingData", "gender")), matchThreshold=0.7)
  objExample5b= amUnique(amDataset(amExample5, indexColumn="sampleId"), matchThreshold=0.7)
  objExample5c= amUnique(amDataset(amExample5, indexColumn="sampleId", metaDataColumn="samplingData"),
                         multilocusMap=c(1,2,2,3,3,4,4,5,5,6,6,7,7,8,8,9,9,10,10,11,11), # "gender" only has one allele in it's locus
                         matchThreshold=0.7)
  sink()

  # The print output changed with regards to spaces in 2.6.0 of allelematch:
  amvariant <- as.character(packageVersion("allelematch")[1, 1:2])

  # Run each of the data sets through the tree ways to print the results:
  withr::local_options(width=200) # Allow longer lines for the summaries:
  for (obj in c("objMini", "objExample5", "objExample5b", "objExample5c")) {

    # Write the name of the amUnique object to the _snap file:
    expect_snapshot(paste("About to exercise", obj), variant=amvariant)

    # allelematch:  Console summary is not available for "amUnique" objects.
    # Please use summary.amUnique(x, html=TRUE) or summary.amUnique(x, csv="file.csv") options.
    # expect_snapshot(summary.amUnique(get(obj)), variant=amvariant)

    # amCSV.amUnique should have the same output as before
    tmp = tempfile(paste(obj, "_", sep=""), fileext=".csv")
    expect_snapshot(summary.amUnique(  get(obj), csv=tmp), variant=amvariant)
    expect_snapshot(format(read.csv(tmp)), variant=amvariant)
    file.remove(tmp)

    # amHTML.amUnique should have the same output as before
    tmp = tempfile(paste(obj, "_", sep=""), fileext=".html")
    expect_snapshot(summary.amUnique( get(obj), html=tmp),
                    variant=amvariant)
    expect_snapshot(
      cat(
        sub("summary generated: </b><em>.+?</em>",
            "summary generated: </b><em>(date)</em>",
            gsub("(\\t| )+?(\\n|$)","\\2",
                readLines(tmp, warn=FALSE),
                perl=TRUE),
            perl=TRUE),
        sep="\n"),
      variant=amvariant
    )
    file.remove(tmp)
  }


  # Test usingTmpFile:
  withr::local_envvar(.new=list(TMP = tempdir()), action="replace")
  {
    # Just test that we can get code coverage for the 'usingTmpFile' code
    # without any errors. The code that generated html from data was tested above.
    out = capture_output(
      amHTML.amUnique( get("objMini"), htmlFile=NULL)
    )
    expect_match(out, "Opening HTML file.+? in default browser", perl=TRUE) # Ignore ever-changing 'tempdir()'
  }

})

#
# This file is called 'tdd-xxxx.R' rather than 'test-xxx.R'.
# It is therefore not included the normal package tests.
# But they can still be executed manually.
#
# The content of a 'tdd-xxxx.R' contains tests that currently fail because of
# some flaws in allelematch.
#
# The intention is to fix the flaws by using the Test Driven Design (TDD) approach.
#
# In TDD, you (1) make sure the flaw can be reproduced and detected with a new test.
# You then (2) fix the flaw and (3) make sure the test are now working.
# You  then (4) move the tests to 'test-xxx.R' to ensure the fixed problems
# never occur again.
#
test_that("Test Driven Design, TDD - Tests that are still failing", {

  # Miniature input sample:
  sample = miniExample = data.frame(
    sampleId        = c(1:4),
    knownIndividual = c("A","A","B","  B"),
    dismiss         = c("Rain", " drops", " keep", " fallin'"),
    "LOC1a"         = c(11:14),
    "LOC1b"         = c(21:24),
    "LOC2a"         = c(31:33, -88),
    "LOC2b"         = c(41:44)
  )

  {
    # These bad parameters should have caused errors:

    # TODO : Catch that integer values for missingCode are stored in amDataset as integer! Should be character!
    # All surprising conditions in the state of an object are error prone.
    x = amDataset(sample, missingCode=-88)

    expect_identical(x$missingCode, "-88")

    # TODO : Catch that indexColumn is overlapped by metaDataColumn or ignoreColumn
    expect_error(amDataset(sample, indexColumn = "sampleId", metaDataColumn = "sampleId"), "unused argument")
    expect_error(amDataset(sample, indexColumn = "sampleId", ignoreColumn = "sampleId"), "unused argument")
  }

})

#' Missing cover:
#' amDataset: Param checking
#'
#' amCluster:
#'    482	489: if (inherits(amDatasetFocal, "amInterpolate")) { ## TODO: Remove? class "amInterpolate" is never set anywhere.
#'    summary.amCluster
#'    amHTML.amCluster
#'    amCSV.amCluster
#'
#' summary.amPairwise
#'    amHTML.amPairwise
#'    amCSV.amPairwise
#'
#' summary.amUnique
#' amUnique : Param checking,
#'    1099--1101
#'
#' amUniqueProfile : Parm checking.
#'   if (verbose)
#' amCSSForHTML
#'

test_that("Validation of arguments to amMatrix() is working", {

  # Test checks against invalid first param to amDataSet(multilocusDataset):
  {
    expect_error(amMatrix(NULL),      "allelematch:\\s+amDatasetFocal must be an object of class \"amDataset\"")
    expect_error(amMatrix(NA),        "allelematch:\\s+amDatasetFocal must be an object of class \"amDataset\"")
  }

  # Create valid miniature input sample:
  miniExample = data.frame(
    # sampleId        = c(1:4),
    # knownIndividual = c("A","A","B","  B"),
    # dismiss         = c("Rain", " drops", " keep", " fallin' on my head "),
    "LOC1a"         = c(11:14),
    "LOC1b"         = c(21:24),
    "LOC2a"         = c(31:33, -99),
    "LOC2b"         = c(41:44)
  )
  {
    expect_snapshot(print(miniExample))
    expect_snapshot(amdataMini <- amDataset(miniExample))
    expect_snapshot(print.amDataset(amdataMini))
  }

  # Use the valid amdata to test blocking of other invalid arguments:
  {
    amdata = amdataMini
    expect_error(amMatrix(amdata, missingNoneExisting =2), "unused argument")
    expect_error(amMatrix(amdata, missingMethod = 0),   "allelematch:\\s+missingMethod must equal 1 or 2")
    expect_error(amMatrix(amdata, missingMethod = 3),   "allelematch:\\s+missingMethod must equal 1 or 2")
    expect_error(amMatrix(amdata, missingMethod = NULL),"argument is of length zero")
    expect_error(amMatrix(amdata, missingMethod = NA),  "allelematch:\\s+missingMethod must equal 1 or 2")
    expect_error(amMatrix(amdata, missingMethod = "2.0"), "allelematch:\\s+missingMethod must equal 1 or 2")
    expect_error(amMatrix(amdata, missingMethod = " 2"), "allelematch:\\s+missingMethod must equal 1 or 2")
    expect_error(amMatrix(amdata, missingMethod = "2 "), "allelematch:\\s+missingMethod must equal 1 or 2")
    expect_error(amMatrix(amdata, missingMethod = 2.000001), "allelematch:\\s+missingMethod must equal 1 or 2")

    expect_snapshot_value(amdata, style = "json2")
  }

  # The following arguments to missingMethod pas. Maybe they shouldn't:
  expect_snapshot(mx0 <- amMatrix(amdataMini, 1) )    # Expected datatype
  {
    # Make sure that different argument types for the parameter "missingMethod"
    # doesn't change the calculation of the amMatrix:
    expect_identical(mx0, amMatrix(amdata, "1"))  # Unexp
    expect_identical(mx0, amMatrix(amdata, 1.0))  # Unexp
    expect_identical(mx0, amMatrix(amdata, 1.0000000000000001))  # Unexp

    expect_snapshot_value(mx0, style = "json2")
  }
})

test_that("Validation of arguments to amDataset() is working", {

  data(amExample1)
  expect_equal(!!dim(amExample1), c(20,22))

  # Test checks against invalid first param to amDataSet(multilocusDataset):
  expect_error(amDataset(NULL), "allelematch:\\s+multilocusDataset must be a matrix or a data.frame")
  expect_error(amDataset("I'm a matrix, prommise!'"), "allelematch:\\s+multilocusDataset must be a matrix or a data.frame")
  expect_error(amDataset(c("I'm", "a", "matrix", "prommise!")), "allelematch:\\s+multilocusDataset must be a matrix or a data.frame")


  # Miniature input sample:
  sample = miniExample = data.frame(
    sampleId        = c(1:4),
    knownIndividual = c("A","A","B","  B"),
    dismiss         = c("Rain", " drops", " keep", " fallin'"),
    "LOC1a"         = c(11:14),
    "LOC1b"         = c(21:24),
    "LOC2a"         = c(31:33, -88),
    "LOC2b"         = c(41:44)
  )

  expect_error(amDataset(sample, .mumboJumbo="-88"), "unused argument")
  expect_error(amDataset(sample, mumboJumbo="-88"), "unused argument")
  expect_error(amDataset(sample, indexColumn    = "None-existing column"), "indexColumn does not exist in multilocusDataset")
  expect_error(amDataset(sample, metaDataColumn = "None-existing column"), "metaDataColumn does not exist in multilocusDataset")
  expect_error(amDataset(sample, ignoreColumn   = "None-existing column"), "one or more ignoreColumn does not exist in multilocusDataset")
  expect_error(amDataset(sample, indexColumn = "knownIndividual"), "index column should contain a unique identifier for each sample")

  # See also currently failing tests in tdd-allelematch_1-amDataset_negative.R
})

#' summary.amCluster()
#'   857-885 : object$cluster
#'
#' amUnique
#'   1099 -- 1101 : if (length(thisLocus)==1) => Gender?
#'
#' amUniqueProfile
#'   1237 -- 1269 : Different calc of matchThreshold, cutHeight, alleleMismatch, profileType.
#'   1351 -- 1361 : Verbose
#'
#' amHTML.* : usingTmpFile
#'
#' amHTML.amPairwise
#'  metadata!
#'
#' amHTML.amCluster
#'  metadata!
#'  1853 -- 1945 ## Clustered genotypes

test_that("Validation of arguments to amUnique() is working", {

  # Create valid miniature input sample:
  miniExample1 = data.frame(
    "LOC1a"         = c(11:14),
    "LOC1b"         = c(21:24),
    "LOC2a"         = c(31:33, -99),
    "LOC2b"         = c(41:44)
  )
  {
    expect_snapshot(print(miniExample1))
    expect_snapshot(amdataMini1 <- amDataset(miniExample1))
    expect_snapshot(print.amDataset(amdataMini1))
  }

  # Parameters are:
  # amUnique <- function(amDatasetFocal, multilocusMap=NULL, alleleMismatch=NULL, matchThreshold=NULL, cutHeight=NULL, doPsib="missing", consensusMethod=1, verbose=FALSE)

  # Test checks against invalid first param to amDataSet(multilocusDataset):
  {
    expect_error(amUnique(NULL),  'allelematch:\\s+amDatasetFocal must be an object of class "amDataset"')
    expect_error(amUnique(NA),    'allelematch:\\s+amDatasetFocal must be an object of class "amDataset"')
    expect_error(amUnique(2),     'allelematch:\\s+amDatasetFocal must be an object of class "amDataset"')
  }

  # Use the valid amdata to test blocking of other invalid arguments:
  amdata = amdataMini1
  sink(nullfile())
  {
    expect_error(amUnique(amdata, multilocusMap = NA),  "allelematch:\\s+multilocusMap must be a vector of integers or strings giving the mappings onto loci")
    expect_error(amUnique(amdata, multilocusMap=c(1,2,3,4,5,6)),  "allelematch:\\s+multilocusMap must be a vector of integers or strings") # Too long multilocusMap
    expect_error(amUnique(amdata, multilocusMap=c(1,1)),  "allelematch:\\s+multilocusMap must be a vector of integers or strings") # Too short multilocusMap
    # expect_error(amUnique(amdata, multilocusMap=c(FALSE,FALSE,TRUE,TRUE)),  "TODO") # TODO: Neither int nor char => Should fail

    expect_error(amUnique(amdata, multilocusMap = c(1,1,2,2)), "allelematch:\\s+please specify alleleMismatch OR matchThreshold OR cutHeight.")

    expect_no_error(amUnique(amdata, multilocusMap = c(1,1,2,2), alleleMismatch=1))
    expect_no_error(amUnique(amdata, alleleMismatch=1))

    oneOf = "^allelematch:\\s+please specify alleleMismatch OR matchThreshold OR cutHeight.$"
    expect_error(amUnique(amdata, alleleMismatch=1, matchThreshold=0.5, cutHeight=0.5), oneOf)
    expect_error(amUnique(amdata, alleleMismatch=1, matchThreshold=0.5), oneOf)
    expect_error(amUnique(amdata, alleleMismatch=1, cutHeight=0.5), oneOf)
    expect_error(amUnique(amdata, matchThreshold=0.5, cutHeight=0.5), oneOf)

    expect_error(amUnique(amdata, alleleMismatch=5)) # TODO: alleleMismatch can't be larger than number of alleles; the specific downstream error is version-dependent
    expect_error(amUnique(amdata, alleleMismatch=-1), "matchThreshold must be between 0 and 1") # TODO: alleleMismatch can't be negative

    expect_error(amUnique(amdata, matchThreshold=-0.0001), "matchThreshold must be between 0 and 1")
    expect_error(amUnique(amdata, matchThreshold=1.0001),  "matchThreshold must be between 0 and 1")
#   expect_error(amUnique(amdata, matchThreshold=0)) # "no clusters formed.\\s+Please set cutHeight lower and run again")  # TODO : Error detected on some platforms, but not all.

    expect_no_error(amUnique(amdata, matchThreshold=1))

    expect_error(amUnique(amdata, cutHeight=-0.0001), "cutHeight must be greater than 0 and less than 1") # TODO: alleleMismatch can't be negative
    expect_error(amUnique(amdata, cutHeight=1.00001), "cutHeight must be greater than 0 and less than 1") # TODO: alleleMismatch can't be negative
    expect_error(amUnique(amdata, cutHeight=1)) # "no clusters formed.\\s+Please set cutHeight lower and run again") # TODO : Error detection and message not alligned

    expect_no_error(amUnique(amdata, cutHeight=0)) # TODO : Error detection and message not alligned
    expect_no_error(amUnique(amdata, cutHeight=0.5))

    expect_no_error(amUnique(amdata, cutHeight=0.5, doPsib="missing"))
    expect_no_error(amUnique(amdata, cutHeight=0.5, doPsib="all"))

    expect_no_error(amUnique(amdata, cutHeight=0.5, doPsib="a")) # TODO : Should fail!
    expect_no_error(amUnique(amdata, cutHeight=0.5, doPsib="")) # TODO : Should fail!
    expect_no_error(amUnique(amdata, cutHeight=0.5, doPsib="THIS SHOULD FAIL!")) # TODO : Should fail!
    expect_error(amUnique(amdata, cutHeight=0.5, doPsib=NULL), "argument is of length zero") # TODO : Bad msg. But perhaps from RLang?
    expect_error(amUnique(amdata, cutHeight=0.5, doPsib=NA), "missing value where TRUE/FALSE needed") # TODO : Bad msg

    expect_no_error(amUnique(amdata, cutHeight=0.5, consensusMethod=1))
    expect_no_error(amUnique(amdata, cutHeight=0.5, consensusMethod=4))

    expect_error(amUnique(amdata, cutHeight=0.5, consensusMethod=0), "consensusMethod must equal 1, 2, 3 or 4")
    expect_error(amUnique(amdata, cutHeight=0.5, consensusMethod=5), "consensusMethod must equal 1, 2, 3 or 4")

    expect_no_error(amUnique(amdata, cutHeight=0.5, verbose=FALSE))

    expect_no_error(amUnique(amdata, cutHeight=0.5, verbose=TRUE))

    expect_error(amUnique(amdata, cutHeight=0.5, verbose="ESLAF"), "argument is not interpretable as logical") # TODO: Bad msg.
  }
  sink()

  # The output below changed between allelematch 2.5.2 and 2.5.3, which is OK:
  expect_output(summary.amUnique(amUnique(amdata, cutHeight=0.5)), 'Console summary is not available for "amUnique" objects.\\s+Please use summary(.amUnique)?\\(x, html=TRUE\\) or summary(.amUnique)?\\(x, csv="file.csv"\\) options.')


})

test_that("Loop the Loop", {
  
  # Run different data sets with different qualities through the same loops:
  miniExample = data.frame(
    "LOC1a"         = c(11:14),
    "LOC1b"         = c(21:24),
    "LOC2a"         = c(31:33, -99),
    "LOC2b"         = c(41:44)
  )
  data("amExample1")
  data("amExample2") ; amExample2 = amExample2[c(1:20),] # Just keep the first 20 rows to save speed and disk
  data("amExample3") ; amExample3 = amExample3[c(1:20),] # Just keep the first 20 rows to save speed and disk
  data("amExample4") ; amExample4 = amExample4[c(1:20),] # Just keep the first 20 rows to save speed and disk
  data("amExample5") ; amExample5 = amExample5[c(1:20),] # Just keep the first 20 rows to save speed and disk
  
  amdataMini     = amDataset(miniExample)
  amdataExample1 = amDataset(amExample1, indexColumn="sampleId", metaDataColumn="knownIndividual")
  amdataExample2 = amDataset(amExample2, indexColumn="sampleId", metaDataColumn="knownIndividual")
  amdataExample3 = amDataset(amExample3, indexColumn="sampleId", metaDataColumn="knownIndividual")
  amdataExample4 = amDataset(amExample4, indexColumn="sampleId", metaDataColumn="knownIndividual")
  amdataExample5 = amDataset(amExample5, indexColumn="sampleId", ignoreColumn=c("samplingData", "gender"))
  
  # multilocusMaps:
  mlm1 = c("LOC1",  "LOC1",  "LOC2", "LOC2", "LOC3", "LOC3", "LOC4", "LOC4", "LOC5", "LOC5", "LOC6", "LOC6", "LOC7", "LOC7", "LOC8", "LOC8", "LOC9", "LOC9", "LOC10", "LOC10")
  mlm2 = c("LOC10", "LOC10", "LOC9", "LOC9", "LOC8", "LOC8", "LOC7", "LOC7", "LOC6", "LOC6", "LOC5", "LOC5", "LOC4", "LOC4", "LOC3", "LOC3", "LOC2", "LOC2", "LOC1",  "LOC1")
  
  
  # We want to run amAlleleFreq many times with many combinations of parameters
  # and we want to compare the results with previous runs. Like this:
  snapshot_amAlleleFreq <- function(ds, multilocusMap=NULL) {
    
    # Log the call to the snapshot file:
    { 
      argstr = ifelse(is.null(multilocusMap), "", paste(", multilocusMap=c(", paste(multilocusMap, collapse=", "), ")", sep=""))
      cmdstr = paste("amAlleleFreq(", ds, argstr, ")", sep="")
      expect_snapshot(cat(cmdstr))
    }

    # Capture any errors reported by allelematch:
    sink(nullfile()) # Block output from 'cat' within allelematch
    af <- tryCatch( 
      
      # Make the call to allelematch:  
      amAlleleFreq(amDatasetFocal=get(ds), multilocusMap=multilocusMap),
      
      ## If the call fails, return the error message and
      ## the method and arguments that threw the error:
      error = function(e) {
        ret = c(paste("\n  Error    : ", e$message, "\n  Rejected : ", cmdstr, "\n"))
        message("\n  ", ret, sep="") # Print the message to the screen for easier debugging
        ret # Return this message from 'tryCatch'
      }
      #, message = function(m) {}
    )
    sink() # Block output from 'cat' within allelematch
    
    # Log the result to the snapshot file:
    # expect_snapshot(summary.amAlleleFreq(af)) # Too big. :-(
    expect_snapshot_value(af, style = "deparse")
    
    if(class(af) == "amAlleleFreq") { 
      expect_snapshot(print.amAlleleFreq(af)) 
    } else {
      expect_snapshot(cat(af))
    }

    return(af)
  }
  
  # Parameters to loop over:
  #   " amAlleleFreq <- function(amDatasetFocal, , multilocusMap=NULL) {
  # Regarding clusterMethod: "Only 'complete' acceptable. This option remains for experimental reasons". So, we stick with the default.
  
  # Here comes the loops:
  af0 = snapshot_amAlleleFreq("amdataMini")
  af1 = snapshot_amAlleleFreq("amdataMini", multilocusMap=c(1,1,2,2))
  expect_identical(af1, af0)

  for (amds in c("amdataExample1", "amdataExample2", "amdataExample3", "amdataExample4", "amdataExample5")) {
    af0 = snapshot_amAlleleFreq(amds)
  }
})

# Validation of arguments to amMatrix() is working

    Code
      print(miniExample)
    Output
        LOC1a LOC1b LOC2a LOC2b
      1    11    21    31    41
      2    12    22    32    42
      3    13    23    33    43
      4    14    24   -99    44

---

    Code
      amdataMini <- amDataset(miniExample)

---

    Code
      print.amDataset(amdataMini)
    Output
      allelematch
      amDataset object
          LOC1a LOC1b LOC2a LOC2b
      AAA    11    21    31    41
      AAB    12    22    32    42
      AAC    13    23    33    43
      AAD    14    24   -99    44

---

    {
      "type": "list",
      "attributes": {
        "names": {
          "type": "character",
          "attributes": {},
          "value": ["index", "multilocus", "missingCode"]
        },
        "class": {
          "type": "character",
          "attributes": {},
          "value": ["amDataset"]
        }
      },
      "value": [
        {
          "type": "character",
          "attributes": {},
          "value": ["AAA", "AAB", "AAC", "AAD"]
        },
        {
          "type": "character",
          "attributes": {
            "dim": {
              "type": "integer",
              "attributes": {},
              "value": [4, 4]
            },
            "dimnames": {
              "type": "list",
              "attributes": {},
              "value": [
                {
                  "type": "NULL"
                },
                {
                  "type": "character",
                  "attributes": {},
                  "value": ["LOC1a", "LOC1b", "LOC2a", "LOC2b"]
                }
              ]
            }
          },
          "value": ["11", "12", "13", "14", "21", "22", "23", "24", "31", "32", "33", "-99", "41", "42", "43", "44"]
        },
        {
          "type": "character",
          "attributes": {},
          "value": ["-99"]
        }
      ]
    }

---

    Code
      mx0 <- amMatrix(amdataMini, 1)

---

    {
      "type": "double",
      "attributes": {
        "dim": {
          "type": "integer",
          "attributes": {},
          "value": [4, 4]
        },
        "dimnames": {
          "type": "list",
          "attributes": {},
          "value": [
            {
              "type": "character",
              "attributes": {},
              "value": ["AAA", "AAB", "AAC", "AAD"]
            },
            {
              "type": "character",
              "attributes": {},
              "value": ["AAA", "AAB", "AAC", "AAD"]
            }
          ]
        },
        "class": {
          "type": "character",
          "attributes": {},
          "value": ["amMatrix"]
        }
      },
      "value": [0, 1, 1, 0.9375, 1, 0, 1, 0.9375, 1, 1, 0, 0.9375, 0.9375, 0.9375, 0.9375, 0.125]
    }

# Loop the Loop

    Code
      cat(cmdstr)
    Output
      amUniqueProfile(amdataMini, doPlot=FALSE, verbose=FALSE)

---

    Code
      cat(ret)
    Output
      allelematch:  assuming genotype columns are in pairs, representing 3 loci
        matchThreshold cutHeight alleleMismatch samples unique unclassified multipleMatch guessOptimum missingDataLoad allelicDiversity guessMorphology
      1      1.0000000 0.0000000              0       4      4            0             0        FALSE           0.083              7.3        ZeroFlat
      2      0.8333333 0.1666667              1       4      4            0             0         TRUE           0.083              7.3        ZeroFlat
      3      0.6666667 0.3333333              2       4      4            0             0        FALSE           0.083              7.3        ZeroFlat

---

    Code
      cat(cmdstr)
    Output
      amUniqueProfile(amdataExample1, doPlot=FALSE, verbose=FALSE)

---

    Code
      cat(ret)
    Output
      allelematch:  assuming genotype columns are in pairs, representing 10 loci
        matchThreshold cutHeight alleleMismatch samples unique unclassified multipleMatch guessOptimum missingDataLoad allelicDiversity   guessMorphology
      1           1.00      0.00              0      20     13            0             0        FALSE           0.005              6.1 ZeroSecondMinimum
      2           0.95      0.05              1      20     13            0             2        FALSE           0.005              6.1 ZeroSecondMinimum
      3           0.90      0.10              2      20     12            0             0         TRUE           0.005              6.1 ZeroSecondMinimum
      4           0.85      0.15              3      20     12            0             0        FALSE           0.005              6.1 ZeroSecondMinimum
      5           0.80      0.20              4      20     12            0             0        FALSE           0.005              6.1 ZeroSecondMinimum
      6           0.75      0.25              5      20     12            0             0        FALSE           0.005              6.1 ZeroSecondMinimum
      7           0.70      0.30              6      20     12            0             0        FALSE           0.005              6.1 ZeroSecondMinimum
      8           0.65      0.35              7      20     12            0             0        FALSE           0.005              6.1 ZeroSecondMinimum
      9           0.60      0.40              8      20     12            0             1        FALSE           0.005              6.1 ZeroSecondMinimum

---

    Code
      cat(cmdstr)
    Output
      amUniqueProfile(amdataExample2, doPlot=FALSE, verbose=FALSE)

---

    Code
      cat(ret)
    Output
      allelematch:  assuming genotype columns are in pairs, representing 10 loci
        matchThreshold cutHeight alleleMismatch samples unique unclassified multipleMatch guessOptimum missingDataLoad allelicDiversity   guessMorphology
      1           1.00      0.00              0     148    140            0             0        FALSE           0.046              7.9 ZeroSecondMinimum
      2           0.95      0.05              1     148    136            0            25        FALSE           0.046              7.9 ZeroSecondMinimum
      3           0.90      0.10              2     148    122            0            33        FALSE           0.046              7.9 ZeroSecondMinimum
      4           0.85      0.15              3     148    100            0             0         TRUE           0.046              7.9 ZeroSecondMinimum
      5           0.80      0.20              4     148     99            0             0        FALSE           0.046              7.9 ZeroSecondMinimum
      6           0.75      0.25              5     148     99            0             1        FALSE           0.046              7.9 ZeroSecondMinimum
      7           0.70      0.30              6     148     98            0             1        FALSE           0.046              7.9 ZeroSecondMinimum
      8           0.65      0.35              7     148     98            0            14        FALSE           0.046              7.9 ZeroSecondMinimum
      9           0.60      0.40              8     148     89            0            30        FALSE           0.046              7.9 ZeroSecondMinimum

---

    Code
      cat(cmdstr)
    Output
      amUniqueProfile(amdataExample3, doPlot=FALSE, verbose=FALSE)

---

    Code
      cat(ret)
    Output
      allelematch:  assuming genotype columns are in pairs, representing 10 loci
        matchThreshold cutHeight alleleMismatch samples unique unclassified multipleMatch guessOptimum missingDataLoad allelicDiversity      guessMorphology
      1           1.00      0.00              0     319    291            0             0        FALSE           0.097              8.2 NonZeroSecondMinimum
      2           0.95      0.05              1     319    281            0            97        FALSE           0.097              8.2 NonZeroSecondMinimum
      3           0.90      0.10              2     319    212            0           103        FALSE           0.097              8.2 NonZeroSecondMinimum
      4           0.85      0.15              3     319    159            0            77        FALSE           0.097              8.2 NonZeroSecondMinimum
      5           0.80      0.20              4     319    120            0            31        FALSE           0.097              8.2 NonZeroSecondMinimum
      6           0.75      0.25              5     319    109            0            25        FALSE           0.097              8.2 NonZeroSecondMinimum
      7           0.70      0.30              6     319    100            2             4         TRUE           0.097              8.2 NonZeroSecondMinimum
      8           0.65      0.35              7     319     93            7            12        FALSE           0.097              8.2 NonZeroSecondMinimum
      9           0.60      0.40              8     319     93            2            75        FALSE           0.097              8.2 NonZeroSecondMinimum

---

    Code
      cat(cmdstr)
    Output
      amUniqueProfile(amdataExample4, doPlot=FALSE, verbose=FALSE)

---

    Code
      cat(ret)
    Output
      allelematch:  assuming genotype columns are in pairs, representing 10 loci
        matchThreshold cutHeight alleleMismatch samples unique unclassified multipleMatch guessOptimum missingDataLoad allelicDiversity guessMorphology
      1           1.00      0.00              0     307    307            0             0        FALSE           0.199              4.8 NoSecondMinimum
      2           0.95      0.05              1     307    307            0            42         TRUE           0.199              4.8 NoSecondMinimum
      3           0.90      0.10              2     307    283            0            92        FALSE           0.199              4.8 NoSecondMinimum
      4           0.85      0.15              3     307    228            1            99        FALSE           0.199              4.8 NoSecondMinimum
      5           0.80      0.20              4     307    180            1           110        FALSE           0.199              4.8 NoSecondMinimum
      6           0.75      0.25              5     307    137            4           122        FALSE           0.199              4.8 NoSecondMinimum
      7           0.70      0.30              6     307     99            1           153        FALSE           0.199              4.8 NoSecondMinimum
      8           0.65      0.35              7     307     66            8           196        FALSE           0.199              4.8 NoSecondMinimum
      9           0.60      0.40              8     307     39           10           234        FALSE           0.199              4.8 NoSecondMinimum

---

    Code
      cat(cmdstr)
    Output
      amUniqueProfile(amdataExample5, doPlot=FALSE, verbose=FALSE)

---

    Code
      cat(ret)
    Output
      allelematch:  assuming genotype columns are in pairs, representing 10 loci
        matchThreshold cutHeight alleleMismatch samples unique unclassified multipleMatch guessOptimum missingDataLoad allelicDiversity      guessMorphology
      1           1.00      0.00              0     335    253            0             0        FALSE           0.064              6.4 NonZeroSecondMinimum
      2           0.95      0.05              1     335    233            0            95        FALSE           0.064              6.4 NonZeroSecondMinimum
      3           0.90      0.10              2     335    189            0            65        FALSE           0.064              6.4 NonZeroSecondMinimum
      4           0.85      0.15              3     335    165            1            32         TRUE           0.064              6.4 NonZeroSecondMinimum
      5           0.80      0.20              4     335    154            0            48        FALSE           0.064              6.4 NonZeroSecondMinimum
      6           0.75      0.25              5     335    136            2           103        FALSE           0.064              6.4 NonZeroSecondMinimum
      7           0.70      0.30              6     335    106            1           163        FALSE           0.064              6.4 NonZeroSecondMinimum
      8           0.65      0.35              7     335     74            3           234        FALSE           0.064              6.4 NonZeroSecondMinimum
      9           0.60      0.40              8     335     40            4           237        FALSE           0.064              6.4 NonZeroSecondMinimum

---

    Code
      print(obj)
    Output
        matchThreshold  cutHeight alleleMismatch samples unique unclassified multipleMatch guessOptimum missingDataLoad allelicDiversity      guessMorphology
      1      1.0000000 0.00000000              0     335    265            0             0        FALSE           0.095              6.1 NonZeroSecondMinimum
      2      0.9523810 0.04761905              1     335    240            0            79        FALSE           0.095              6.1 NonZeroSecondMinimum
      3      0.9047619 0.09523810              2     335    195            1            57        FALSE           0.095              6.1 NonZeroSecondMinimum
      4      0.8571429 0.14285714              3     335    171            3            11         TRUE           0.095              6.1 NonZeroSecondMinimum
      5      0.8095238 0.19047619              4     335    155            1            41        FALSE           0.095              6.1 NonZeroSecondMinimum
      6      0.7619048 0.23809524              5     335    141            1            62        FALSE           0.095              6.1 NonZeroSecondMinimum
      7      0.7142857 0.28571429              6     335    115            2           123        FALSE           0.095              6.1 NonZeroSecondMinimum
      8      0.6666667 0.33333333              7     335     84           17            70        FALSE           0.095              6.1 NonZeroSecondMinimum
      9      0.6190476 0.38095238              8     335     45           26           176        FALSE           0.095              6.1 NonZeroSecondMinimum

# See how an object of class amDataset is built:

    Code
      print(miniExample)
    Output
        sampleId knownIndividual             dismiss. LOC1a LOC1b LOC2a LOC2b
      1        1               A                 Rain    11    21    31    41
      2        2               A                drops    12    22    32    42
      3        3               B                 keep    13    23    33    43
      4        4               C  fallin' on my head     14    24   -88    44

---

    Code
      miniDataset1 <- amDataset(miniExample)

---

    Code
      print.amDataset(miniDataset1)
    Output
      allelematch
      amDataset object
          sampleId knownIndividual        dismiss. LOC1a LOC1b LOC2a LOC2b
      AAA        1               A            Rain    11    21    31    41
      AAB        2               A           drops    12    22    32    42
      AAC        3               B            keep    13    23    33    43
      AAD        4               C fallin'onmyhead    14    24   -88    44

---

    {
      "type": "list",
      "attributes": {
        "names": {
          "type": "character",
          "attributes": {},
          "value": ["index", "multilocus", "missingCode"]
        },
        "class": {
          "type": "character",
          "attributes": {},
          "value": ["amDataset"]
        }
      },
      "value": [
        {
          "type": "character",
          "attributes": {},
          "value": ["AAA", "AAB", "AAC", "AAD"]
        },
        {
          "type": "character",
          "attributes": {
            "dim": {
              "type": "integer",
              "attributes": {},
              "value": [4, 7]
            },
            "dimnames": {
              "type": "list",
              "attributes": {},
              "value": [
                {
                  "type": "NULL"
                },
                {
                  "type": "character",
                  "attributes": {},
                  "value": ["sampleId", "knownIndividual", "dismiss.", "LOC1a", "LOC1b", "LOC2a", "LOC2b"]
                }
              ]
            }
          },
          "value": ["1", "2", "3", "4", "A", "A", "B", "C", "Rain", "drops", "keep", "fallin'onmyhead", "11", "12", "13", "14", "21", "22", "23", "24", "31", "32", "33", "-88", "41", "42", "43", "44"]
        },
        {
          "type": "character",
          "attributes": {},
          "value": ["-99"]
        }
      ]
    }

---

    Code
      miniDataset2 <- amDataset(miniExample, missingCode = "-88", indexColumn = "sampleId",
        metaDataColumn = "knownIndividual", ignoreColumn = "dismiss.")

---

    Code
      print.amDataset(miniDataset2)
    Output
      allelematch
      amDataset object
             LOC1a LOC1b LOC2a LOC2b
      1  A      11    21    31    41
      2  A      12    22    32    42
      3  B      13    23    33    43
      4    C    14    24   -88    44

---

    {
      "type": "list",
      "attributes": {
        "names": {
          "type": "character",
          "attributes": {},
          "value": ["index", "metaData", "multilocus", "missingCode"]
        },
        "class": {
          "type": "character",
          "attributes": {},
          "value": ["amDataset"]
        }
      },
      "value": [
        {
          "type": "character",
          "attributes": {},
          "value": ["1", "2", "3", "4"]
        },
        {
          "type": "character",
          "attributes": {},
          "value": ["A", "A", "B", "  C"]
        },
        {
          "type": "character",
          "attributes": {
            "dim": {
              "type": "integer",
              "attributes": {},
              "value": [4, 4]
            },
            "dimnames": {
              "type": "list",
              "attributes": {},
              "value": [
                {
                  "type": "NULL"
                },
                {
                  "type": "character",
                  "attributes": {},
                  "value": ["LOC1a", "LOC1b", "LOC2a", "LOC2b"]
                }
              ]
            }
          },
          "value": ["11", "12", "13", "14", "21", "22", "23", "24", "31", "32", "33", "-88", "41", "42", "43", "44"]
        },
        {
          "type": "character",
          "attributes": {},
          "value": ["-88"]
        }
      ]
    }

---

    Code
      miniDataset3 <- amDataset(miniExample, missingCode = "-88", indexColumn = 1,
        metaDataColumn = 2, ignoreColumn = 3)

---

    Code
      print.amDataset(miniDataset2)
    Output
      allelematch
      amDataset object
             LOC1a LOC1b LOC2a LOC2b
      1  A      11    21    31    41
      2  A      12    22    32    42
      3  B      13    23    33    43
      4    C    14    24   -88    44

---

    {
      "type": "list",
      "attributes": {
        "names": {
          "type": "character",
          "attributes": {},
          "value": ["index", "metaData", "multilocus", "missingCode"]
        },
        "class": {
          "type": "character",
          "attributes": {},
          "value": ["amDataset"]
        }
      },
      "value": [
        {
          "type": "character",
          "attributes": {},
          "value": ["1", "2", "3", "4"]
        },
        {
          "type": "character",
          "attributes": {},
          "value": ["A", "A", "B", "  C"]
        },
        {
          "type": "character",
          "attributes": {
            "dim": {
              "type": "integer",
              "attributes": {},
              "value": [4, 4]
            },
            "dimnames": {
              "type": "list",
              "attributes": {},
              "value": [
                {
                  "type": "NULL"
                },
                {
                  "type": "character",
                  "attributes": {},
                  "value": ["LOC1a", "LOC1b", "LOC2a", "LOC2b"]
                }
              ]
            }
          },
          "value": ["11", "12", "13", "14", "21", "22", "23", "24", "31", "32", "33", "-88", "41", "42", "43", "44"]
        },
        {
          "type": "character",
          "attributes": {},
          "value": ["-88"]
        }
      ]
    }

# Different data types for arg to 'missingCode' give same result

    Code
      ds1 <- amDataset(sample, missingCode = "NA")
    Output
      allelematch:  NA data converted to NA 

---

    structure(list(index = c("AAA", "AAB", "AAC", "AAD"), multilocus = structure(c("11", 
    "12", "13", "14", "21", "22", "23", "NA", "31", "32", "33", "NA", 
    "41", "42", "43", "NA"), dim = c(4L, 4L), dimnames = list(NULL, 
        c("LOC1a", "LOC1b", "LOC2a", "LOC2b"))), missingCode = "NA"), class = "amDataset")

---

    Code
      ds2 <- amDataset(sample, missingCode = NA)
    Output
      allelematch:  NA data converted to NA 

---

    structure(list(index = c("AAA", "AAB", "AAC", "AAD"), multilocus = structure(c("11", 
    "12", "13", "14", "21", "22", "23", "NA", "31", "32", "33", "NA", 
    "41", "42", "43", "NA"), dim = c(4L, 4L), dimnames = list(NULL, 
        c("LOC1a", "LOC1b", "LOC2a", "LOC2b"))), missingCode = "NA"), class = "amDataset")

# Validation of arguments to amUnique() is working

    Code
      print(miniExample1)
    Output
        LOC1a LOC1b LOC2a LOC2b
      1    11    21    31    41
      2    12    22    32    42
      3    13    23    33    43
      4    14    24   -99    44

---

    Code
      amdataMini1 <- amDataset(miniExample1)

---

    Code
      print.amDataset(amdataMini1)
    Output
      allelematch
      amDataset object
          LOC1a LOC1b LOC2a LOC2b
      AAA    11    21    31    41
      AAB    12    22    32    42
      AAC    13    23    33    43
      AAD    14    24   -99    44

# Validation of arguments to amPairwise() is working

    Code
      print(miniExample1)
    Output
        LOC1a LOC1b LOC2a LOC2b
      1    11    21    31    41
      2    12    22    32    42
      3    13    23    33    43
      4    14    24   -99    44

---

    Code
      amdataMini1 <- amDataset(miniExample1)

---

    Code
      print.amDataset(amdataMini1)
    Output
      allelematch
      amDataset object
          LOC1a LOC1b LOC2a LOC2b
      AAA    11    21    31    41
      AAB    12    22    32    42
      AAC    13    23    33    43
      AAD    14    24   -99    44

---

    {
      "type": "list",
      "attributes": {
        "names": {
          "type": "character",
          "attributes": {},
          "value": ["index", "multilocus", "missingCode"]
        },
        "class": {
          "type": "character",
          "attributes": {},
          "value": ["amDataset"]
        }
      },
      "value": [
        {
          "type": "character",
          "attributes": {},
          "value": ["AAA", "AAB", "AAC", "AAD"]
        },
        {
          "type": "character",
          "attributes": {
            "dim": {
              "type": "integer",
              "attributes": {},
              "value": [4, 4]
            },
            "dimnames": {
              "type": "list",
              "attributes": {},
              "value": [
                {
                  "type": "NULL"
                },
                {
                  "type": "character",
                  "attributes": {},
                  "value": ["LOC1a", "LOC1b", "LOC2a", "LOC2b"]
                }
              ]
            }
          },
          "value": ["11", "12", "13", "14", "21", "22", "23", "24", "31", "32", "33", "-99", "41", "42", "43", "44"]
        },
        {
          "type": "character",
          "attributes": {},
          "value": ["-99"]
        }
      ]
    }

---

    Code
      print(oddExample2)
    Output
        LOC1a LOC1b LOC2a LOC2b LOC3a
      1    11    21    31    41    51
      2    12    22    32    42    52
      3    13    23    33    43    53
      4    14    24   -99    44   -99

---

    Code
      amdataOdd2 <- amDataset(oddExample2)

---

    Code
      print.amDataset(amdataOdd2)
    Output
      allelematch
      amDataset object
          LOC1a LOC1b LOC2a LOC2b LOC3a
      AAA    11    21    31    41    51
      AAB    12    22    32    42    52
      AAC    13    23    33    43    53
      AAD    14    24   -99    44   -99

---

    {
      "type": "list",
      "attributes": {
        "names": {
          "type": "character",
          "attributes": {},
          "value": ["index", "multilocus", "missingCode"]
        },
        "class": {
          "type": "character",
          "attributes": {},
          "value": ["amDataset"]
        }
      },
      "value": [
        {
          "type": "character",
          "attributes": {},
          "value": ["AAA", "AAB", "AAC", "AAD"]
        },
        {
          "type": "character",
          "attributes": {
            "dim": {
              "type": "integer",
              "attributes": {},
              "value": [4, 5]
            },
            "dimnames": {
              "type": "list",
              "attributes": {},
              "value": [
                {
                  "type": "NULL"
                },
                {
                  "type": "character",
                  "attributes": {},
                  "value": ["LOC1a", "LOC1b", "LOC2a", "LOC2b", "LOC3a"]
                }
              ]
            }
          },
          "value": ["11", "12", "13", "14", "21", "22", "23", "24", "31", "32", "33", "-99", "41", "42", "43", "44", "51", "52", "53", "-99"]
        },
        {
          "type": "character",
          "attributes": {},
          "value": ["-99"]
        }
      ]
    }

---

    {
      "type": "list",
      "attributes": {
        "names": {
          "type": "character",
          "attributes": {},
          "value": ["index", "multilocus", "missingCode"]
        },
        "class": {
          "type": "character",
          "attributes": {},
          "value": ["amDataset"]
        }
      },
      "value": [
        {
          "type": "character",
          "attributes": {},
          "value": ["AAA", "AAB", "AAC", "AAD"]
        },
        {
          "type": "character",
          "attributes": {
            "dim": {
              "type": "integer",
              "attributes": {},
              "value": [4, 4]
            },
            "dimnames": {
              "type": "list",
              "attributes": {},
              "value": [
                {
                  "type": "NULL"
                },
                {
                  "type": "character",
                  "attributes": {},
                  "value": ["LOC1a", "LOC1b", "LOC2a", "LOC2b"]
                }
              ]
            }
          },
          "value": ["11", "12", "13", "14", "21", "22", "23", "24", "31", "32", "33", "-99", "41", "42", "43", "44"]
        },
        {
          "type": "character",
          "attributes": {},
          "value": ["-99"]
        }
      ]
    }

---

    Code
      pw0 <- amPairwise(amdata, alleleMismatch = 2)

---

    {
      "type": "list",
      "attributes": {
        "names": {
          "type": "character",
          "attributes": {},
          "value": ["pairwise", "missingCode", "matchThreshold", "alleleMismatch", "missingMethod", "focalDatasetN", "comparisonDatasetN", "focalIsComparison"]
        },
        "class": {
          "type": "character",
          "attributes": {},
          "value": ["amPairwise"]
        }
      },
      "value": [
        {
          "type": "list",
          "attributes": {},
          "value": [
            {
              "type": "list",
              "attributes": {
                "names": {
                  "type": "character",
                  "attributes": {},
                  "value": ["focal", "match"]
                }
              },
              "value": [
                {
                  "type": "list",
                  "attributes": {
                    "names": {
                      "type": "character",
                      "attributes": {},
                      "value": ["index", "metaData", "multilocus", "flags"]
                    }
                  },
                  "value": [
                    {
                      "type": "character",
                      "attributes": {},
                      "value": ["AAA"]
                    },
                    {
                      "type": "NULL"
                    },
                    {
                      "type": "character",
                      "attributes": {
                        "dim": {
                          "type": "integer",
                          "attributes": {},
                          "value": [1, 4]
                        },
                        "dimnames": {
                          "type": "list",
                          "attributes": {},
                          "value": [
                            {
                              "type": "NULL"
                            },
                            {
                              "type": "character",
                              "attributes": {},
                              "value": ["LOC1a", "LOC1b", "LOC2a", "LOC2b"]
                            }
                          ]
                        }
                      },
                      "value": ["11", "21", "31", "41"]
                    },
                    {
                      "type": "double",
                      "attributes": {
                        "dim": {
                          "type": "integer",
                          "attributes": {},
                          "value": [1, 4]
                        }
                      },
                      "value": [1, 1, 1, 1]
                    }
                  ]
                },
                {
                  "type": "list",
                  "attributes": {
                    "names": {
                      "type": "character",
                      "attributes": {},
                      "value": ["index", "multilocus", "score", "flags", "perfect", "partial"]
                    }
                  },
                  "value": [
                    {
                      "type": "character",
                      "attributes": {},
                      "value": ["AAA"]
                    },
                    {
                      "type": "character",
                      "attributes": {
                        "dim": {
                          "type": "integer",
                          "attributes": {},
                          "value": [1, 4]
                        },
                        "dimnames": {
                          "type": "list",
                          "attributes": {},
                          "value": [
                            {
                              "type": "NULL"
                            },
                            {
                              "type": "character",
                              "attributes": {},
                              "value": ["LOC1a", "LOC1b", "LOC2a", "LOC2b"]
                            }
                          ]
                        }
                      },
                      "value": ["11", "21", "31", "41"]
                    },
                    {
                      "type": "character",
                      "attributes": {},
                      "value": ["1"]
                    },
                    {
                      "type": "double",
                      "attributes": {
                        "dim": {
                          "type": "integer",
                          "attributes": {},
                          "value": [1, 4]
                        }
                      },
                      "value": [1, 1, 1, 1]
                    },
                    {
                      "type": "integer",
                      "attributes": {},
                      "value": [1]
                    },
                    {
                      "type": "integer",
                      "attributes": {},
                      "value": [0]
                    }
                  ]
                }
              ]
            },
            {
              "type": "list",
              "attributes": {
                "names": {
                  "type": "character",
                  "attributes": {},
                  "value": ["focal", "match"]
                }
              },
              "value": [
                {
                  "type": "list",
                  "attributes": {
                    "names": {
                      "type": "character",
                      "attributes": {},
                      "value": ["index", "metaData", "multilocus", "flags"]
                    }
                  },
                  "value": [
                    {
                      "type": "character",
                      "attributes": {},
                      "value": ["AAB"]
                    },
                    {
                      "type": "NULL"
                    },
                    {
                      "type": "character",
                      "attributes": {
                        "dim": {
                          "type": "integer",
                          "attributes": {},
                          "value": [1, 4]
                        },
                        "dimnames": {
                          "type": "list",
                          "attributes": {},
                          "value": [
                            {
                              "type": "NULL"
                            },
                            {
                              "type": "character",
                              "attributes": {},
                              "value": ["LOC1a", "LOC1b", "LOC2a", "LOC2b"]
                            }
                          ]
                        }
                      },
                      "value": ["12", "22", "32", "42"]
                    },
                    {
                      "type": "double",
                      "attributes": {
                        "dim": {
                          "type": "integer",
                          "attributes": {},
                          "value": [1, 4]
                        }
                      },
                      "value": [1, 1, 1, 1]
                    }
                  ]
                },
                {
                  "type": "list",
                  "attributes": {
                    "names": {
                      "type": "character",
                      "attributes": {},
                      "value": ["index", "multilocus", "score", "flags", "perfect", "partial"]
                    }
                  },
                  "value": [
                    {
                      "type": "character",
                      "attributes": {},
                      "value": ["AAB"]
                    },
                    {
                      "type": "character",
                      "attributes": {
                        "dim": {
                          "type": "integer",
                          "attributes": {},
                          "value": [1, 4]
                        },
                        "dimnames": {
                          "type": "list",
                          "attributes": {},
                          "value": [
                            {
                              "type": "NULL"
                            },
                            {
                              "type": "character",
                              "attributes": {},
                              "value": ["LOC1a", "LOC1b", "LOC2a", "LOC2b"]
                            }
                          ]
                        }
                      },
                      "value": ["12", "22", "32", "42"]
                    },
                    {
                      "type": "character",
                      "attributes": {},
                      "value": ["1"]
                    },
                    {
                      "type": "double",
                      "attributes": {
                        "dim": {
                          "type": "integer",
                          "attributes": {},
                          "value": [1, 4]
                        }
                      },
                      "value": [1, 1, 1, 1]
                    },
                    {
                      "type": "integer",
                      "attributes": {},
                      "value": [1]
                    },
                    {
                      "type": "integer",
                      "attributes": {},
                      "value": [0]
                    }
                  ]
                }
              ]
            },
            {
              "type": "list",
              "attributes": {
                "names": {
                  "type": "character",
                  "attributes": {},
                  "value": ["focal", "match"]
                }
              },
              "value": [
                {
                  "type": "list",
                  "attributes": {
                    "names": {
                      "type": "character",
                      "attributes": {},
                      "value": ["index", "metaData", "multilocus", "flags"]
                    }
                  },
                  "value": [
                    {
                      "type": "character",
                      "attributes": {},
                      "value": ["AAC"]
                    },
                    {
                      "type": "NULL"
                    },
                    {
                      "type": "character",
                      "attributes": {
                        "dim": {
                          "type": "integer",
                          "attributes": {},
                          "value": [1, 4]
                        },
                        "dimnames": {
                          "type": "list",
                          "attributes": {},
                          "value": [
                            {
                              "type": "NULL"
                            },
                            {
                              "type": "character",
                              "attributes": {},
                              "value": ["LOC1a", "LOC1b", "LOC2a", "LOC2b"]
                            }
                          ]
                        }
                      },
                      "value": ["13", "23", "33", "43"]
                    },
                    {
                      "type": "double",
                      "attributes": {
                        "dim": {
                          "type": "integer",
                          "attributes": {},
                          "value": [1, 4]
                        }
                      },
                      "value": [1, 1, 1, 1]
                    }
                  ]
                },
                {
                  "type": "list",
                  "attributes": {
                    "names": {
                      "type": "character",
                      "attributes": {},
                      "value": ["index", "multilocus", "score", "flags", "perfect", "partial"]
                    }
                  },
                  "value": [
                    {
                      "type": "character",
                      "attributes": {},
                      "value": ["AAC"]
                    },
                    {
                      "type": "character",
                      "attributes": {
                        "dim": {
                          "type": "integer",
                          "attributes": {},
                          "value": [1, 4]
                        },
                        "dimnames": {
                          "type": "list",
                          "attributes": {},
                          "value": [
                            {
                              "type": "NULL"
                            },
                            {
                              "type": "character",
                              "attributes": {},
                              "value": ["LOC1a", "LOC1b", "LOC2a", "LOC2b"]
                            }
                          ]
                        }
                      },
                      "value": ["13", "23", "33", "43"]
                    },
                    {
                      "type": "character",
                      "attributes": {},
                      "value": ["1"]
                    },
                    {
                      "type": "double",
                      "attributes": {
                        "dim": {
                          "type": "integer",
                          "attributes": {},
                          "value": [1, 4]
                        }
                      },
                      "value": [1, 1, 1, 1]
                    },
                    {
                      "type": "integer",
                      "attributes": {},
                      "value": [1]
                    },
                    {
                      "type": "integer",
                      "attributes": {},
                      "value": [0]
                    }
                  ]
                }
              ]
            },
            {
              "type": "list",
              "attributes": {
                "names": {
                  "type": "character",
                  "attributes": {},
                  "value": ["focal", "match"]
                }
              },
              "value": [
                {
                  "type": "list",
                  "attributes": {
                    "names": {
                      "type": "character",
                      "attributes": {},
                      "value": ["index", "metaData", "multilocus", "flags"]
                    }
                  },
                  "value": [
                    {
                      "type": "character",
                      "attributes": {},
                      "value": ["AAD"]
                    },
                    {
                      "type": "NULL"
                    },
                    {
                      "type": "character",
                      "attributes": {
                        "dim": {
                          "type": "integer",
                          "attributes": {},
                          "value": [1, 4]
                        },
                        "dimnames": {
                          "type": "list",
                          "attributes": {},
                          "value": [
                            {
                              "type": "NULL"
                            },
                            {
                              "type": "character",
                              "attributes": {},
                              "value": ["LOC1a", "LOC1b", "LOC2a", "LOC2b"]
                            }
                          ]
                        }
                      },
                      "value": ["14", "24", "-99", "44"]
                    },
                    {
                      "type": "double",
                      "attributes": {
                        "dim": {
                          "type": "integer",
                          "attributes": {},
                          "value": [1, 4]
                        }
                      },
                      "value": [1, 1, 2, 1]
                    }
                  ]
                },
                {
                  "type": "list",
                  "attributes": {
                    "names": {
                      "type": "character",
                      "attributes": {},
                      "value": ["index", "multilocus", "score", "flags", "perfect", "partial"]
                    }
                  },
                  "value": [
                    {
                      "type": "character",
                      "attributes": {},
                      "value": ["AAD"]
                    },
                    {
                      "type": "character",
                      "attributes": {
                        "dim": {
                          "type": "integer",
                          "attributes": {},
                          "value": [1, 4]
                        },
                        "dimnames": {
                          "type": "list",
                          "attributes": {},
                          "value": [
                            {
                              "type": "NULL"
                            },
                            {
                              "type": "character",
                              "attributes": {},
                              "value": ["LOC1a", "LOC1b", "LOC2a", "LOC2b"]
                            }
                          ]
                        }
                      },
                      "value": ["14", "24", "-99", "44"]
                    },
                    {
                      "type": "character",
                      "attributes": {},
                      "value": ["1"]
                    },
                    {
                      "type": "double",
                      "attributes": {
                        "dim": {
                          "type": "integer",
                          "attributes": {},
                          "value": [1, 4]
                        }
                      },
                      "value": [1, 1, 2, 1]
                    },
                    {
                      "type": "integer",
                      "attributes": {},
                      "value": [1]
                    },
                    {
                      "type": "integer",
                      "attributes": {},
                      "value": [0]
                    }
                  ]
                }
              ]
            }
          ]
        },
        {
          "type": "character",
          "attributes": {},
          "value": ["-99"]
        },
        {
          "type": "double",
          "attributes": {},
          "value": [0.5]
        },
        {
          "type": "double",
          "attributes": {},
          "value": [2]
        },
        {
          "type": "double",
          "attributes": {},
          "value": [2]
        },
        {
          "type": "integer",
          "attributes": {},
          "value": [4]
        },
        {
          "type": "integer",
          "attributes": {},
          "value": [4]
        },
        {
          "type": "logical",
          "attributes": {},
          "value": [true]
        }
      ]
    }

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/amregtest-package.R
\name{amregtest}
\alias{amregtest}
\title{Package Overview}
\description{
Package 'amregtest' automates regression testing of package \link[allelematch:allelematch-package]{allelematch}.

The API is simple. There are only three functions:
\tabular{clcl}{
\verb{  } \tab \code{\link{artRun}()}\verb{  }\tab \verb{  } \tab Executes the test, or a subset of the tests\cr
\verb{  } \tab \code{\link{artList}()}\tab \tab Lists the available tests without running them\cr
\verb{  } \tab \code{\link{artVersion}()}\tab \tab Shows the used versions of 'allelematch' and 'amregtest'\cr
}

The prefix "art" is short for "Allelematch Regression Test".

See \link{artData} for a description of data sets used as input.
}
\references{
\url{https://github.com/cran/allelematch}

\href{https://github.com/cran/allelematch/blob/2.5.1/inst/doc/allelematchSuppDoc.pdf}{allelematchSuppDoc.pdf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/art_api.R
\name{artRun}
\alias{artRun}
\title{Runs the regression test}
\usage{
artRun(filter = "", verbose = TRUE)
}
\arguments{
\item{filter}{If specified, only tests with names matching this perl regular
expression will be executed. Character vector of length 1. See also \link{artList}}

\item{verbose}{logical. If TRUE (the default), prints version of tested allelematch to stdout}
}
\value{
A list (invisibly) containing data about the test results as returned by \link[testthat:test_package]{testthat::test_package}
}
\description{
Runs regression tests on package \link[allelematch:allelematch-package]{allelematch} to make sure it is backwards compatible.\cr
\cr
The full set of tests will take a couple of minutes. \cr
\cr
Call \link{artList} to see the available tests without running them.
}
\details{
If any of the test executed with \link{artRun} should fail, then we want to be able
to run that specific test under the debugger.\cr
\cr
Set a breakpoint in \code{allelematch.R} and call \code{artRun(filter="<the test that reproduces the problem>")}\cr
\cr
Note that it is the last loaded version of \code{allelematch} that will be executed,
not the last edited. In RStudio, CTRL+SHIFT+B will build, install and load.
}
\examples{
# See what version of packages 'allelematch' and 'amregtest'
# are currently loaded:
artVersion()

# List the available tests:
artList()
\donttest{
# Run all the tests:
# artRun()  # Takes several minutes

# Run the first of the available tests:
artRun(filter="allelematch_1-amDataset$")
}

}
\seealso{
\link{artVersion} and \link{artList}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/art_data.R
\docType{data}
\name{artData}
\alias{artData}
\title{Example data used by \link{amregtest}}
\format{
Data frames with varying numbers of samples in rows, and alleles in columns. Missing data is represented as "-99".
}
\description{
This example data is used when testing allelematch backwards compatibility
using \link{artRun}. The tests load this data and passes it to \link[allelematch:amDataset]{amDataset}.\cr
\cr
It includes data that was imported from version 5.2.1 of \link[allelematch:allelematch-package]{allelematch}.
It was still unchanged in 5.2.5. \cr
\tabular{clcl}{
\verb{  } \tab \link{amExample1}\verb{  }\tab \verb{  } \tab Example 1 High quality data set\cr
\verb{  } \tab \link{amExample2}\tab \tab Example 2 Good quality data set\cr
\verb{  } \tab \link{amExample3}\tab \tab Example 3 Marginal quality data set\cr
\verb{  } \tab \link{amExample4}\tab \tab Example 4 Low quality data set\cr
\verb{  } \tab \link{amExample5}\tab \tab Example 5 Wildlife data set\cr
}
See \href{https://github.com/cran/allelematch/blob/2.5.1/inst/doc/allelematchSuppDoc.pdf}{allelematchSuppDoc.pdf}
for a more detailed description. \cr

It also includes a large data set gathered from field work:\cr
\tabular{clcl}{
\verb{  } \tab \link{ggSample}\verb{  }\tab \verb{  } \tab Very large wildlife data set\cr
}
}
\references{
\url{https://github.com/cran/allelematch}

\href{https://github.com/cran/allelematch/blob/2.5.1/inst/doc/allelematchSuppDoc.pdf}{allelematchSuppDoc.pdf}
}
\keyword{data}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/art_data.R
\docType{data}
\name{amExample3}
\alias{amExample3}
\title{Example 3 Marginal quality data set}
\format{
Data frame with samples in rows, and alleles in columns. Missing data is represented as "-99".
}
\description{
This is sample data copied from \link[allelematch:amExampleData]{amExampleData} in version 5.2.1
of package \link[allelematch:allelematch-package]{allelematch}.
}
\details{
The data in this example have been simulated to represent a data set of marginal
quality where the use of allelematch combined with careful manual review of the
results is required to achieve a confident assessment of the unique genotypes. At each
locus a random 4\% of heterozygotes lost their second allele to simulate an allele dropout,
and a random 10\% of samples at each locus had alleles set to missing.
}
\references{
\url{https://github.com/cran/allelematch}
}
\keyword{data}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/art_data.R
\docType{data}
\name{amExample1}
\alias{amExample1}
\title{Example 1 High quality data set}
\format{
Data frame with samples in rows, and alleles in columns. Missing data is represented as "-99".
}
\description{
This is sample data copied from \link[allelematch:amExampleData]{amExampleData} in version 5.2.1
of package \link[allelematch:allelematch-package]{allelematch}. We use this data to test 'allelematch' backwards compatibility.
}
\details{
The data in this example is simulated to represent a high quality data set that might
result from a laboratory protocol where samples were run multiple times to confirm their
identity. It has no genotyping error, a near-zero missing data load, and approximately
60\% of the individuals have been artificially resampled more than once.
}
\references{
\url{https://github.com/cran/allelematch}
}
\keyword{data}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/art_data.R
\docType{data}
\name{ggSample}
\alias{ggSample}
\title{Data sets originating from GG work}
\format{
This data is saved on semicolon (";") delimited .csv format, as described under 'Details' in \link[utils:data]{utils::data}.\cr
Data samples in rows, and alleles in columns. Missing data is represented as "-99".
}
\description{
Large data set gathered from field work in 2022. Here used to test package \link[allelematch:allelematch-package]{allelematch} for backwards compatibility.
Combines a reference db of known individuals with new samples to be analyzed.\cr\cr
}
\references{
\url{https://github.com/cran/allelematch}
}
\keyword{data}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/art_api.R
\name{artList}
\alias{artList}
\title{Lists available tests in \code{amregtest} without running them}
\usage{
artList(verbose = TRUE)
}
\arguments{
\item{verbose}{logical. If TRUE (the default), prints additional info to stdout}
}
\value{
A character vector containing the names of all the tests
}
\description{
Use the output to select a value for parameter \code{filter} to \link{artRun}.
Useful when debugging.
}
\examples{
# See what version of packages 'allelematch' and 'amregtest'
# are currently loaded:
artVersion()

# List the available tests:
artList()
\donttest{
# Run all the tests:
# artRun()  # Takes several minutes

# Run the first of the available tests:
artRun(filter="allelematch_1-amDataset$")
}

}
\seealso{
\link{artVersion} and \link{artRun}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/art_data.R
\docType{data}
\name{amExample5}
\alias{amExample5}
\title{Example 5 Wildlife data set}
\format{
Data frame with samples in rows, and alleles in columns. Missing data is represented as "-99".
}
\description{
This is sample data copied from \link[allelematch:amExampleData]{amExampleData} in version 5.2.1
of package \link[allelematch:allelematch-package]{allelematch}.
}
\details{
In this final example we use real data from the non-invasive sampling of a wildlife
population. The data have been anonymized by changing sampling details. A single
column giving the gender is also available and we show how this can be used as an extra
locus. Missing data is also more common at some loci than at others, with a total load
of about 10\%.
}
\references{
\url{https://github.com/cran/allelematch}
}
\keyword{data}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/amregtest-package.R
\docType{package}
\name{amregtest-package}
\alias{amregtest-package}
\title{amregtest: Runs Allelematch Regression Tests}
\description{
Automates regression testing of package 'allelematch'. Over 2500 tests cover all functions in 'allelematch', reproduce the examples from the documentation, and include negative tests. The implementation is based on 'testthat'.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/torstax/amregtest}
  \item Report bugs at \url{https://github.com/torstax/amregtest/issues}
}

}
\author{
\strong{Maintainer}: Torvald Staxler \email{torvald.staxler@telia.com}

Authors:
\itemize{
  \item Torvald Staxler \email{torvald.staxler@telia.com}
}

Other contributors:
\itemize{
  \item Department of Wildlife, Fish and Environmental Studies at Swedish University of Agricultural Sciences [copyright holder]
  \item Göran Spong \email{Goran.Spong@slu.se} (Senior Lecturer at the Department of Wildlife, Fish and Environmental Studies) [copyright holder]
  \item Paul Galpern \email{pgalpern@gmail.com} (Author of the excellent package 'allelematch' from which the five amExample data files have been copied under MIT license) [contributor]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/art_data.R
\docType{data}
\name{amExample2}
\alias{amExample2}
\title{Example 2 Good quality data set}
\format{
Data frame with samples in rows, and alleles in columns. Missing data is represented as "-99".
}
\description{
This is sample data copied from \link[allelematch:amExampleData]{amExampleData} in version 5.2.1
of package \link[allelematch:allelematch-package]{allelematch}.
}
\details{
The data in this example have also been simulated, this time to reflect the qualities
of good quality data set, where genotyping error and missing data exist, but these can
be confidently handled by allelematch without manual intervention. At each locus a
random 4\% of heterozygotes lost their second allele to simulate an allele dropout, and
a random 4\% of samples at each locus had alleles set to missing.
}
\references{
\url{https://github.com/cran/allelematch}
}
\keyword{data}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/art_data.R
\docType{data}
\name{amExample4}
\alias{amExample4}
\title{Example 4 Low quality data set}
\format{
Data frame with samples in rows, and alleles in columns. Missing data is represented as "-99".
}
\description{
This is sample data copied from \link[allelematch:amExampleData]{amExampleData} in version 5.2.1
of package \link[allelematch:allelematch-package]{allelematch}.
}
\details{
For this example we have simulated a low quality data set where uncertainty created
by genotyping error and missing data, combined with a lack of information in the form
of allelic diversity across loci will result in a low confidence assessment of the unique
genotypes. At each locus a random 6\% of heterozygotes lost their second allele to
simulate an allele dropout, and a random 20\% of samples at each locus had alleles set
to missing.
}
\references{
\url{https://github.com/cran/allelematch}
}
\keyword{data}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/art_api.R
\name{artVersion}
\alias{artVersion}
\title{Returns package version}
\usage{
artVersion(verbose = TRUE)
}
\arguments{
\item{verbose}{logical. If TRUE (the default), prints additional info to stdout,
including versions and build timestamps of 'allelematch' and 'amregtest'.}
}
\value{
The loaded version of this package (\link{amregtest-package}) in a character vector of length one
}
\description{
Displays version of this package (\link{amregtest}) and of \link[allelematch:allelematch-package]{allelematch},
together with build timestamps.\cr
\cr
The version is specified in the file DESCRIPTION, tag "Version: ".
}
\examples{
# See what version of packages 'allelematch' and 'amregtest'
# are currently loaded:
artVersion()

# List the available tests:
artList()
\donttest{
# Run all the tests:
# artRun()  # Takes several minutes

# Run the first of the available tests:
artRun(filter="allelematch_1-amDataset$")
}


}
\seealso{
\link{artList}, \link{artRun} and \link{amregtest}
}
